import jax
import jax.numpy as jnp


def generate_column_index_matrix(num_rows, num_cols):
    """
    Create a 2D array of indexes used for the sorting technique.

    This code is based on the method described in:
    [Nghia T. Vo et al., 2018] - "Superior techniques for eliminating ring artifacts in x-ray micro-tomography"

    This code is adapted from the Tomopy library:
    https://github.com/tomopy/tomopy.git

    References:
    [1] Vo N, and Atwood RC, and Drakopoulos M. Superior techniques for eliminating ring artifacts in x-ray micro-tomography. Optics Express, 26(22):28396–28412, 2018.
    [2] Tomopy library https://github.com/tomopy/tomopy.git

    Args:
        num_rows (int): number of detector rows in the sinogram
        num_cols (int): number of detector channels in the sinogram

    Returns:
        index_matrix(jax array): a 2D jax array of indexes
    """
    list_index = jnp.arange(0.0, num_cols, 1.0)
    index_matrix = jnp.tile(list_index, (num_rows, 1))
    return index_matrix


def remove_small_stripes_sorting(sino, filter_size, index_matrix):
    """
    Remove small-to-medium partial and fulll stripes using the sorting technique.

    This code is based on the method described in:
    [Nghia T. Vo et al., 2018] - "Superior techniques for eliminating ring artifacts in x-ray micro-tomography"

    This code is adapted from the Tomopy library:
    https://github.com/tomopy/tomopy.git

    References:
    [1] Vo N, and Atwood RC, and Drakopoulos M. Superior techniques for eliminating ring artifacts in x-ray micro-tomography. Optics Express, 26(22):28396–28412, 2018.
    [2] Tomopy library https://github.com/tomopy/tomopy.git

    Args:
        sino (jax array): a 2D slice of the sinogram data with shape (num_views, num_det_channels)
        filter_size (int): window size of the median filter
        index_matrix (jax array): a 2D array of indexes used for the sorting technique

    Return:
        corrected_sino (jax array): corrected 2D slice of the sinogram data after stripes removal
    """

    from scipy.ndimage import median_filter
    # Sort each column of the sinogram by its grayscale values
    sino = jnp.transpose(sino)
    stacked_matrix = jnp.stack([index_matrix, sino], axis=2)
    sorted_indices = jnp.argsort(stacked_matrix[:, :, 1], axis=1)
    sorted_indices_expanded = sorted_indices[:, :, None]
    sorted_stacked_matrix = jnp.take_along_axis(stacked_matrix, sorted_indices_expanded, axis=1)

    # Apply the median filter on teh sorted sinogram along each row
    sorted_stacked_matrix = sorted_stacked_matrix.at[:, :, 1].set(median_filter(sorted_stacked_matrix[:, :, 1], (filter_size, 1)))

    # Re-sort the smoothed image columns to the original rows to get the corrected sinogram
    sorted_indices = jnp.argsort(sorted_stacked_matrix[:, :, 0], axis=1)
    sorted_indices_expanded = sorted_indices[:, :, None]
    sort_back_matrix = jnp.take_along_axis(sorted_stacked_matrix, sorted_indices_expanded, axis=1)
    corrected_sino = sort_back_matrix[:, :, 1]
    corrected_sino = jnp.transpose(corrected_sino)

    return corrected_sino

def detect_stripe(list_data, snr):
    """
    Used to locate stripes.
    A segmentation algorithm to separate the extremely positive and negative defects from the normal values in the
    sinogram.

    This code is based on the method described in:
    [Nghia T. Vo et al., 2018] - "Superior techniques for eliminating ring artifacts in x-ray micro-tomography"

    This code is adapted from the Tomopy library:
    https://github.com/tomopy/tomopy.git

    Args:
        list_data (jax array): a normalized 1D array
        snr (float): a ratio between the defective value and the background value. a reasonable choice of snr should be around 3.0 or above

    Returns:
        list_mask (jax array): a 2D binary array denoting the stripes detected
    """

    num_data = list_data.shape[0]

    # Sort the 1D array
    sorted_list = jnp.sort(list_data)[::-1]
    x_list = jnp.arange(0, num_data, 1.0)

    # Apply a linear fit to values around the middle of the sorted array
    # Calculating the noise level to avoid false positives caused by minor background variations
    num_data_drop = jnp.int16(0.25 * num_data)
    (_slope, _intercept) = jnp.polyfit(x_list[num_data_drop:-num_data_drop - 1], sorted_list[num_data_drop:-num_data_drop - 1], 1)
    fitted_value_last_index = _intercept + _slope * x_list[-1]
    noise_level = jnp.abs(fitted_value_last_index - _intercept)
    noise_level = jnp.clip(noise_level, 1e-6, None)
    val1 = jnp.abs(sorted_list[0] - _intercept) / noise_level
    val2 = jnp.abs(sorted_list[-1] - fitted_value_last_index) / noise_level

    # Calculate the upper threshold and the lower threshold
    # Binarize the array by replacing all values between lower and upper threshold with 0 and others with 1.
    list_mask = jnp.zeros_like(list_data)
    if (val1 >= snr):
        upper_threshold = _intercept + noise_level * snr * 0.5
        list_mask = jnp.where(list_data > upper_threshold, 1.0, list_mask)
    if (val2 >= snr):
        lower_threshold = fitted_value_last_index - noise_level * snr * 0.5
        list_mask = jnp.where(list_data <= lower_threshold, 1.0, list_mask)

    return list_mask


def remove_large_stripes_sorting(sino, snr, filter_size, index_matrix, drop_ratio=0.1):
    """
    Remove large partial and full stripes using the sorting technique.

    This code is based on the method described in:
    [Nghia T. Vo et al., 2018] - "Superior techniques for eliminating ring artifacts in x-ray micro-tomography"

    This code is adapted from the Tomopy library:
    https://github.com/tomopy/tomopy.git

    References:
    [1] Vo N, and Atwood RC, and Drakopoulos M. Superior techniques for eliminating ring artifacts in x-ray micro-tomography. Optics Express, 26(22):28396–28412, 2018.
    [2] Tomopy library https://github.com/tomopy/tomopy.git

    Args:
        sino (jax array): a 2D slice of the sinogram data with shape (num_views, num_det_channels)
        snr (float): a ratio between the defective value and the background value
        filter_size (int): window size of the median filter
        index_matrix (jnp array): a 2D array of indexes used for the sorting technique
        drop_ratio (float, optional): ratio of pixels at the top and the bottom of the sinogram to be removed. Defaults to 0.1.

    Returns:
        sino (jax array): corrected 2D slice of the sinogram data after stripes removal
    """

    from scipy.ndimage import median_filter, binary_dilation
    drop_ratio = jnp.clip(drop_ratio, 0.0, 0.8)
    (num_rows, num_cols) = sino.shape
    num_rows_drop = jnp.int16(0.5 * drop_ratio * num_rows)

    # Sorting the columns of the sinogram and apply the median filter on the sorted image along each row
    sorted_sino = jnp.sort(sino, axis=0)
    smoothed_sino = jnp.array(median_filter(sorted_sino, (1, filter_size)))

    # Compute the column-wise average of the sorted and smoothed sinogram
    # Compute the normalized 1D array
    raw_list = jnp.mean(sorted_sino[num_rows_drop:num_rows - num_rows_drop], axis=0)
    smoothed_list = jnp.mean(smoothed_sino[num_rows_drop:num_rows - num_rows_drop], axis=0)
    normalized_list = jnp.where(
        smoothed_list != 0,
        jnp.divide(raw_list, smoothed_list),
        jnp.ones_like(raw_list)
    )

    # Locate the large stripes
    list_mask = detect_stripe(normalized_list, snr)
    list_mask = jnp.array(binary_dilation(list_mask, iterations=1).astype(list_mask.dtype))
    normalized_factor = jnp.tile(normalized_list, (num_rows, 1))

    # Apply pre-correction to the original sinogram
    sino = sino / normalized_factor

    # Apply the sorting-based algorithm again to get the corrected columns
    sino_T = jnp.transpose(sino)
    stacked_matrix = jnp.stack([index_matrix, sino_T], axis=2)

    sorted_indices = jnp.argsort(stacked_matrix[:, :, 1], axis=1)
    sorted_indices_expanded = sorted_indices[:, :, None]
    sorted_matrix = jnp.take_along_axis(stacked_matrix, sorted_indices_expanded, axis=1)

    sorted_matrix = sorted_matrix.at[:, :, 1].set(jnp.transpose(smoothed_sino))

    sorted_indices = jnp.argsort(sorted_matrix[:, :, 0], axis=1)
    sorted_indices_expanded = sorted_indices[:, :, None]
    sort_back_matrix = jnp.take_along_axis(sorted_matrix, sorted_indices_expanded, axis=1)

    corrected_sino = jnp.transpose(sort_back_matrix[:, :, 1])

    list_x_miss = jnp.where(list_mask > 0.0)[0]

    # Selective Replacement of Defective Columns with corrected columns
    sino = sino.at[:, list_x_miss].set(corrected_sino[:, list_x_miss])
    return sino


def remove_dead_fluctuating_stripes_interpolation(sino, snr, filter_size, index_matrix):
    """
    Remove unresponsive and fluctuating stripes using the interpolation technique.
    Sorting approach does not work here because the rankings of the grayscales are significantly different between
    pixels inside the stripes and outside the stripes. Instead, interpolation is an appropriate choice.

    This code is based on the method described in:
    [Nghia T. Vo et al., 2018] - "Superior techniques for eliminating ring artifacts in x-ray micro-tomography"

    This code is adapted from the Tomopy library:
    https://github.com/tomopy/tomopy.git

    References:
    [1] Vo N, and Atwood RC, and Drakopoulos M. Superior techniques for eliminating ring artifacts in x-ray micro-tomography. Optics Express, 26(22):28396–28412, 2018.
    [2] Tomopy library https://github.com/tomopy/tomopy.git

    Args:
        sino (jax array): a 2D slice of the sinogram data with shape (num_views, num_det_channels)
        snr (float): a ratio between the defective value and the background value
        filter_size (int): window size of the median filter
        index_matrix (jax array): a 2D array of indexes used for the sorting technique

    Returns:
        sino (jax array): corrected 2D slice of the sinogram data after stripes removal

    """
    from scipy import interpolate
    from scipy.ndimage import uniform_filter1d, median_filter, binary_dilation
    num_rows = sino.shape[0]

    # Compute the column-wise absolute difference between the original sinogram and the smoothed sinogram
    # Help detect the unresponsive and fluctuating stripes (large diff -> fluctuating stripes; small diff -> unresponsive stripes)
    smoothed_sino = jnp.array(uniform_filter1d(sino, 10, axis=0))
    difference_list = jnp.sum(jnp.abs(sino - smoothed_sino), axis=0)

    # Compute normalized 1D array where the large value correspond to defective pixels
    difference_list_filtered = jnp.array(median_filter(difference_list, size=filter_size))
    normalized_list = jnp.where(
        difference_list_filtered != 0,
        jnp.divide(difference_list, difference_list_filtered),
        jnp.ones_like(difference_list_filtered)
    )

    # Generate binary mask
    list_mask = detect_stripe(normalized_list, snr)
    list_mask = jnp.array(binary_dilation(list_mask, iterations=1).astype(list_mask.dtype))
    list_mask = list_mask.at[0:2].set(0.0)
    list_mask = list_mask.at[-2:].set(0.0)

    # Interpolation
    x_list = jnp.where(list_mask < 1.0)[0]
    y_list = jnp.arange(num_rows)
    z_matrix = sino[:, x_list]
    fit_function = interpolate.RectBivariateSpline(y_list, x_list, z_matrix, kx=1, ky=1)

    # Apply interpolation to defective columns
    list_x_miss = jnp.where(list_mask > 0.0)[0]
    if len(list_x_miss) > 0:
        matrix_x_miss, matrix_y = jnp.meshgrid(list_x_miss, y_list)
        estimate_output = fit_function.ev(jnp.ravel(matrix_y), jnp.ravel(matrix_x_miss))
        sino = sino.at[:, list_x_miss].set(jnp.reshape(estimate_output, matrix_x_miss.shape))

    # Remove residual large stripes
    corrected_sino = remove_large_stripes_sorting(sino, snr, filter_size, index_matrix)

    return corrected_sino

def remove_all_stripe(sino, snr=3, large_filter_size=61, small_filter_size=21):
    """
    Removes all types of stripe artifacts from a sinogram using a combination of three algorithms:
    1. Interpolation-based removal of unresponsive and fluctuating stripes.
    2. Sorting-based removal of large partial and full stripes.
    3. Sorting-based removal of small to medium partial and full stripes.

    This method is adapted from `tomopy.remove_all_stripes()` and is based on:
    Vo N, Atwood RC, Drakopoulos M. "Superior techniques for eliminating ring artifacts in x-ray micro-tomography."
    Optics Express, 26(22):28396–28412, 2018.

    Args:
        sino (jax.Array): A 3D sinogram array with shape (num_views, num_det_rows, num_det_channels).
        snr (float, optional): Signal-to-noise ratio used for stripe detection. A typical value is 3.0. Defaults to 3.
        large_filter_size (int, optional): Median filter window size for removing large stripes. Defaults to 61.
        small_filter_size (int, optional): Median filter window size for removing small-to-medium stripes. Defaults to 21.

    Returns:
        jax.Array: Corrected 3D sinogram array after removing all stripe artifacts.

    Example:
        >>> import jax.numpy as jnp
        >>> import mbirjax.preprocess as mjp
        >>> sino = jnp.ones((180, 128, 256))  # Simulated 3D sinogram
        >>> cleaned_sino = mjp.remove_all_stripe(sino)
    """
    from concurrent.futures import ThreadPoolExecutor
    index_matrix = generate_column_index_matrix(sino.shape[2], sino.shape[0])
    index_matrix_cpu = jax.device_put(index_matrix, device=jax.devices("cpu")[0])

    sino_cpu = jax.device_put(sino, device=jax.devices("cpu")[0])

    result = jnp.zeros_like(sino)

    def process_slice(m):
        sino_slice = sino_cpu[:, m, :]
        sino_slice = remove_dead_fluctuating_stripes_interpolation(sino_slice, snr, large_filter_size, index_matrix_cpu)
        sino_slice = remove_small_stripes_sorting(sino_slice, small_filter_size, index_matrix_cpu)
        return sino_slice

    with ThreadPoolExecutor() as executor:
        processed_slices = list(executor.map(process_slice, range(sino.shape[1])))

    for m, processed_slice in enumerate(processed_slices):
        result = result.at[:, m, :].set(processed_slice)

    return jax.device_put(result)


def remove_stripe_fw(sino, wavelet_filter_name="db5", sigma=2):
    """
    Removes vertical stripe artifacts from a 3D sinogram using a combined wavelet-Fourier filtering technique.

    This method uses a 2D Discrete Wavelet Transform followed by a 2D Fourier transform to suppress vertical stripes,
    as described in:
    Beat Münch et al., "Stripe and ring artifact removal with combined wavelet—Fourier filtering", Optics Express, 2009.

    This implementation is adapted from the Tomopy library's `remove_stripe_fw()`:
    https://github.com/tomopy/tomopy.git

    Args:
        sino (jax.Array): 3D sinogram data with shape (num_views, num_det_rows, num_det_channels).
        wavelet_filter_name (str, optional): Wavelet filter type (e.g., 'db5', 'haar'). Defaults to 'db5'.
        sigma (float, optional): Damping parameter in the Fourier domain. Controls the strength of stripe suppression.
            Defaults to 2.

    Returns:
        jax.Array: Corrected sinogram data with reduced vertical stripe artifacts.

    Example:
        >>> import jax.numpy as jnp
        >>> import mbirjax.preprocess as mjp
        >>> sino = jnp.ones((180, 128, 256))  # Simulated sinogram
        >>> cleaned_sino = mjp.remove_stripe_fw(sino)
    """
    # Determine decomposition level L
    import pywt
    level = int(jnp.ceil(jnp.log2(jnp.max(jnp.array(sino.shape)))))
    views, num_rows, num_columns = sino.shape
    padded_views = views + views // 8
    shift_val = views // 16

    for m in range(sino.shape[1]):
        sino_slice = jnp.zeros((padded_views, num_columns), dtype=jnp.float32)
        sino_slice = sino_slice.at[shift_val:views + shift_val].set(sino[:, m, :])

        # 2D Discrete Wavelte Transform
        cH, cV, cD = {}, {}, {}
        for n in range(level):
            sino_slice, (cH[n], cV[n], cD[n]) = pywt.dwt2(sino_slice, wavelet_filter_name)

        # FFT transform of horizontal frequency bands
        for n in range(level):
            # FFT
            fcV = jnp.fft.fftshift(jnp.fft.fft(cV[n], axis=0))
            my, mx = fcV.shape

            # Damping of vertical stripe information
            y_hat = (jnp.arange(-my, my, 2, dtype='float32') + 1) / 2
            damp = -jnp.expm1(-jnp.square(y_hat) / (2 * jnp.square(sigma)))
            fcV *= jnp.transpose(jnp.tile(damp, (mx, 1)))

            # Inverse FFT
            cV[n] = jnp.real(jnp.fft.ifft(jnp.fft.ifftshift(fcV), axis=0))

        # 2D inverse discrete wavelet transform
        for n in range(level)[::-1]:
            sino_slice = sino_slice[0:cH[n].shape[0], 0:cH[n].shape[1]]
            sino_slice = pywt.idwt2((sino_slice, (cH[n], cV[n], cD[n])), wavelet_filter_name)

        sino = sino.at[:, m, :].set(sino_slice[shift_val:views + shift_val, 0:num_columns])

    return sino

def remove_sino_offset(sino):
    """
    Remove additive offsets in the sinogram caused by material outside the field of view.

    This function corrects each row of the sinogram so that the sum over channels is constant
    across views and equal to the minimum sum observed across all views.

    Args:
        sino (jax.Array): Sinogram with shape (num_views, num_rows, num_channels).

    Returns:
        jax.Array: Corrected sinogram with the same shape as the input.

    Example:
        >>> import jax.numpy as jnp
        >>> import mbirjax.preprocess as mjp
        >>> sino = jnp.ones((180, 128, 256)) + jnp.linspace(0, 1, 180)[:, None, None]
        >>> corrected_sino = mjp.remove_sino_offset(sino)
    """
    # Compute the average over channels: shape [view, row]
    sino_channel_avg = jnp.mean(sino, axis=2)

    # Compute the minimum of the channel average across views: shape [row]
    sino_min_channel_avg = jnp.min(sino_channel_avg, axis=0)

    # Compute corrected sinogram
    sino_corrected = sino - sino_channel_avg[:, :, None] + sino_min_channel_avg[None, :, None]

    return sino_corrected




import numpy as np
import jax.numpy as jnp
import matplotlib.pyplot as plt
import mbirjax as mj
from scipy.sparse.linalg import svds, eigsh, aslinearoperator, LinearOperator
import jax

import mbirjax as mj

if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune the vcd reconstruction with a parallel beam projector
    """
    view_batch_size = None
    pixel_batch_size = None
    with jax.experimental.enable_x64(True):  # Finite difference requires 64 bit arithmetic

        # Initialize sinogram
        num_views = 32
        num_det_rows = 1
        num_det_channels = 32
        start_angle = 0
        end_angle = jnp.pi
        sinogram = jnp.zeros((num_views, num_det_rows, num_det_channels))
        angles = jnp.linspace(start_angle, jnp.pi, num_views, endpoint=False)

        # Set up parallel beam model
        parallel_model = mj.ParallelBeamModel(sinogram.shape, angles)

        # Generate phantom
        recon_shape = parallel_model.get_params('recon_shape')
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
        phantom = mj.generate_3d_shepp_logan_low_dynamic_range((num_det_channels, num_det_channels, num_det_channels))
        phantom = phantom[:, :, num_det_channels // 2]

        # Generate indices of pixels
        num_subsets = 4
        full_indices = mj.gen_pixel_partition(recon_shape, num_subsets)[0]

        # Generate sinogram data
        voxel_values = phantom.reshape((-1,) + recon_shape[2:])[full_indices]

        print('Starting forward projection')
        sinogram = parallel_model.sparse_forward_project(voxel_values, full_indices)

        # Determine resulting number of views, slices, and channels and image size
        print('Sinogram shape: {}'.format(sinogram.shape))

        # Get the vector of indices
        all_indices = jnp.arange(num_recon_rows * num_recon_cols)

        sinogram = jnp.array(sinogram)
        all_indices = jnp.array(all_indices)

        hess = parallel_model.compute_hessian_diagonal().flatten()

        # Run once to finish compiling
        print('Starting back projection')
        bp = parallel_model.sparse_back_project(sinogram, all_indices)
        print('Recon shape: ({}, {}, {})'.format(num_recon_rows, num_recon_cols, num_recon_slices))

        input_size = all_indices.size
        input_shape = (all_indices.size, 1)
        output_size = sinogram.size
        output_shape = sinogram.shape

        # Set up for svd
        def Ax_full(local_x):
            local_x = np.reshape(local_x, input_shape)
            ax_flat = parallel_model.sparse_forward_project(local_x, all_indices)
            ax_flat = np.array(ax_flat.flatten())
            return ax_flat

        def Aty_full(local_y):
            local_y = np.reshape(local_y, output_shape)
            aty_flat = parallel_model.sparse_back_project(local_y, all_indices)
            aty_flat = np.array(aty_flat.flatten())
            return aty_flat

        def precond_AtAx(local_x):
            atax = Aty_full(Ax_full(local_x))
            precond_atax = atax / hess
            return precond_atax

        def precond_AtAx_T(local_x):
            local_x = local_x / hess
            atax = Aty_full(Ax_full(local_x))
            return atax

        AtAx_linear_operator = LinearOperator(matvec=precond_AtAx, rmatvec=precond_AtAx_T, shape=(input_size, input_size))

        operator_shape = (sinogram.size, input_size)
        num_sing_values = np.amin(operator_shape) - 20
        eig_vects = True
        print('Computing full AtAx / H eigen-decomposition')
        if eig_vects:
            u, s, vh = svds(AtAx_linear_operator, k=num_sing_values, tol=1e-6, return_singular_vectors=True, solver='propack')
            vh = vh[::-1, :]
            u = u[:, ::-1]
            # mj.slice_viewer(vh.reshape(num_sing_values, num_det_channels, num_det_channels), slice_axis=0)
            # mj.slice_viewer(u.reshape((num_det_channels, num_det_channels, num_sing_values)), slice_axis=2)
        else:
            s = svds(AtAx_linear_operator, k=num_sing_values, tol=1e-6, return_singular_vectors=False, solver='propack')

        s = s[::-1]

        # Get a mask
        g = 1 / num_subsets
        mask = np.array(np.random.rand(*phantom.shape) < g)
        mask = mask.reshape((-1, 1))
        mask_indices = np.where(mask)[0]
        hess_m = hess[mask_indices]

        # Define the masked operators for svd
        def Ax_masked(local_x):
            local_x = local_x.reshape((-1, 1))
            ax_flat = parallel_model.sparse_forward_project(local_x, mask_indices)
            ax_flat = np.array(ax_flat.flatten())
            return ax_flat

        def Aty_masked(local_y):
            local_y = local_y.reshape(output_shape)
            aty_flat = parallel_model.sparse_back_project(local_y, mask_indices)
            aty_flat = np.array(aty_flat.flatten())
            return aty_flat

        def precond_AtAx_masked(local_x):
            atax = Aty_masked(Ax_masked(local_x))
            precond_atax = atax / hess_m
            return precond_atax

        def precond_AtAx_T_masked(local_x):
            local_x = local_x / hess_m
            atax = Aty_masked(Ax_masked(local_x))
            return atax

        # Get the svd for the masked operator
        masked_operator_shape = (len(mask_indices), len(mask_indices))
        Ax_masked_linear_operator = LinearOperator(matvec=precond_AtAx_masked, rmatvec=precond_AtAx_T_masked, shape=masked_operator_shape)

        print('Computing masked AtA / H eigen-decomposition')
        num_masked_sing_values = np.amin(masked_operator_shape) - 1
        num_masked_sing_values = np.minimum(num_masked_sing_values, num_sing_values)
        u_m, s_m, vh_m = svds(Ax_masked_linear_operator, k=num_masked_sing_values, tol=1e-6, return_singular_vectors=True, solver='propack')
        vh_m = vh_m[::-1, :]
        u_m = u_m[:, ::-1]
        s_m = s_m[::-1]

        vm = np.zeros((vh_m.shape[0], np.prod(phantom.shape)))
        vm[:, mask_indices] = vh_m
        vm = vm.reshape((vm.shape[0],) + phantom.shape)
        mj.slice_viewer(vm, slice_axis=0, title='Eigenimages for masked A matrix, {} subsets'.format(num_subsets))

        fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))
        plt.suptitle('recon_shape={}; masking with {} subsets'.format(recon_shape, num_subsets))

        axs[0, 0].semilogy(s, '.')
        axs[0, 0].set_title('Singular values of AtA / H')
        axs[0, 1].semilogy(s_m, '.')
        axs[0, 1].set_title('Singular values of masked AtA / H')
        y_limits = list(axs[0, 0].get_ylim())
        y_limits[0] = 1e-2
        axs[0, 0].set_ylim(y_limits)
        axs[0, 1].set_ylim(y_limits)

        im0 = axs[1, 0].imshow(u @ (np.diag(s) @ vh))
        fig.colorbar(im0, ax=axs[1, 0])
        axs[1, 0].set_title('Full AtA / H')

        im1 = axs[1, 1].imshow(u_m @ (np.diag(s_m) @ vh_m))
        fig.colorbar(im1, ax=axs[1, 1])
        axs[1, 1].set_title('Masked AtA / H')
        plt.show()
        a = 0


import os
import numpy as np
import jax.numpy as jnp
import jax
import time
import matplotlib.pyplot as plt
import gc
import mbirjax as mj
import vcd_coimg_utils

if __name__ == "__main__":
    """
    Investigate the psf
    """
    # ##########################
    # Do all the setup

    # Initialize sinogram
    num_views = 128
    num_det_rows = 20
    num_det_channels = 128
    start_angle = 0
    end_angle = jnp.pi
    sinogram = jnp.zeros((num_views, num_det_rows, num_det_channels))
    angles = jnp.linspace(start_angle, jnp.pi, num_views, endpoint=False)

    # Initialize a random key
    seed_value = np.random.randint(1000000)
    key = jax.random.PRNGKey(seed_value)

    # Set up parallel beam model
    # parallel_model = mj.ParallelBeamModel.from_file('params_parallel.yaml')
    parallel_model = mj.ParallelBeamModel(sinogram.shape, angles)
    # parallel_model.to_file('params_parallel.yaml')

    # Generate phantom
    recon_shape = parallel_model.get_params('recon_shape')
    num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
    phantom = mj.gen_cube_phantom(recon_shape)

    # Generate indices of pixels
    num_subsets = 1
    full_indices = mj.gen_pixel_partition(recon_shape, num_subsets)
    num_subsets = 5
    subset_indices = mj.gen_pixel_partition(recon_shape, num_subsets)

    # Get the vector of indices
    indices = jnp.arange(num_recon_rows * num_recon_cols)
    num_trials = 3
    indices = jnp.mod(np.arange(num_trials, dtype=int).reshape((-1, 1)) + indices.reshape((1, -1)), num_recon_rows * num_recon_cols)

    sinogram = jnp.array(sinogram)
    indices = jnp.array(indices)

    # ##########################
    # Show the forward and back projection from a single pixel
    i, j = num_recon_rows // 2, num_recon_cols // 2
    x = jnp.zeros(recon_shape)
    x = x.at[i, j, :].set(1)
    voxel_values = x.reshape((-1, num_recon_slices))[indices[0]]

    Ax = parallel_model.sparse_forward_project(voxel_values, indices[0])
    Aty = parallel_model.sparse_back_project(Ax, indices[0])
    Aty = parallel_model.reshape_recon(Aty)
    slice_index = num_recon_slices // 2
    Aty_normalized = Aty / Aty[i, j, slice_index]
    reference = 0.45 / np.clip(np.abs(np.arange(num_recon_rows) - (num_recon_cols / 2)), 0.45, None)

    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
    ax[0].imshow(np.log10(Aty_normalized[:, :, slice_index]))
    ax[0].set_title('log10(AtAx)')
    ax[1].semilogy(Aty_normalized[:, num_recon_cols // 2, 0])
    ax[1].semilogy(reference)
    ax[1].legend(['AtAx / max, restricted to a line', 'C / |v - v0|, clipped'])
    ax[1].set_ylim(top=10)
    ax[1].set_title('Slice of AtAx')
    fig_folder = mj.make_figure_folder()
    plt.savefig(os.path.join(fig_folder, 'AtAx_normalized.png'), bbox_inches='tight')
    plt.show()

    a = 0



# -*- coding: utf-8 -*-
"""demo_5_fbp_fdk.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/10ZiCSk1C9D4Fb7Uv6jTtQYxF2lKjmbyh

**MBIRJAX: FBP and FDK Reconstruction Demo**

See the [MBIRJAX documentation](https://mbirjax.readthedocs.io/en/latest/) for an overview and details.  

This script demonstrates the MBIRJAX code by creating a 3D phantom inspired by Shepp-Logan, forward projecting it to create a sinogram, and then using MBIRJAX to perform Filtered Back Projection (FBP) for parallel beam reconstruction and Feldkamp-Davis-Kress reconstruction (FDK) for cone beam reconstruntion.

For the demo, we create some synthetic data by first making a phantom, then forward projecting it to obtain a sinogram.  

In a real application, you would load your sinogram as a numpy array and use numpy.transpose if needed so that it
has axes in the order (views, rows, channels).  For reference, assuming the rotation axis is vertical, then increasing the row index nominally moves down the rotation axis and increasing the channel index moves to the right as seen from the source.

Select a GPU as runtime type for best performance.
"""

# Commented out IPython magic to ensure Python compatibility.
# %pip install -i https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple mbirjax

import numpy as np
import time
import mbirjax as mj

"""**Set the geometry parameters**"""

# Choose the geometry type
geometry_type = 'parallel'  # 'cone' or 'parallel'

# Set parameters for the problem size - you can vary these, but if you make num_det_rows very small relative to
# channels, then the generated phantom may not have an interior.
num_views = 128
num_det_rows = 128
num_det_channels = 128

"""**Data generation:** For demo purposes, we create a phantom and then project it to create a sinogram.

Note:  the sliders on the viewer won't work in notebook form.  For that you'll need to run the python code with an interactive matplotlib backend, typcially using the command line or a development environment like Spyder or Pycharm to invoke python.  

"""

# Choose the geometry type
model_type = 'parallel'  # 'cone' or 'parallel'
object_type = 'shepp-logan'  # 'shepp-logan' or 'cube'

# Generate simulated data
# In a real application you would not have the phantom, but we include it here for later display purposes
phantom, sinogram, params = mj.generate_demo_data(object_type=object_type, model_type=model_type,
                                                  num_views=num_views, num_det_rows=num_det_rows,
                                                  num_det_channels=num_det_channels)
angles = params['angles']

# View sinogram
title = 'Original sinogram \nUse the sliders to change the view or adjust the intensity range.'
mj.slice_viewer(sinogram, slice_axis=0, title=title, slice_label='View')

"""**Initialize for the reconstruction**"""

# ####################
# Use the parameters to get the data and initialize the model for reconstruction.
if model_type == 'cone':
    source_detector_dist = params['source_detector_dist']
    source_iso_dist = params['source_iso_dist']
    ct_model = mj.ConeBeamModel(sinogram.shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
else:
    ct_model = mj.ParallelBeamModel(sinogram.shape, angles)

# Print out model parameters
ct_model.print_params()

"""**Do the reconstruction and display the results.**"""

# ##########################
# Perform FBP/FDK reconstruction
time0 = time.time()
recon = ct_model.direct_recon(sinogram, filter_name="ramp")

recon.block_until_ready()
elapsed = time.time() - time0
# ##########################

max_diff = np.amax(np.abs(phantom - recon))
print('Geometry = {}'.format(geometry_type))
nrmse = np.linalg.norm(recon - phantom) / np.linalg.norm(phantom)
pct_95 = np.percentile(np.abs(recon - phantom), 95)
print('NRMSE between recon and phantom = {}'.format(nrmse))
print('Maximum pixel difference between phantom and recon = {}'.format(max_diff))
print('95% of recon pixels are within {} of phantom'.format(pct_95))

mj.get_memory_stats()
print('Elapsed time for recon is {:.3f} seconds'.format(elapsed))

# Display results
title = (f"Phantom (left) vs {'FDK' if geometry_type == 'cone' else 'FBP'} Recon (right). "
         f"Filter used: ramp. \nUse the sliders to change the slice or adjust the intensity range.")

recon_dict = ct_model.get_recon_dict()
mj.slice_viewer(phantom, recon, data_dicts=[None, recon_dict], title=title)

"""**Next:** Try changing some of the parameters and re-running or try [some of the other demos](https://mbirjax.readthedocs.io/en/latest/demos_and_faqs.html).  """


import numpy as np
import mbirjax as mj
import unittest


class TestProx(unittest.TestCase):
    """
    Test the proximal map function for the special case when the input and initial condition are both the ground truth.
    """

    def setUp(self):
        """Set up before each test method."""
        np.random.seed(0)  # Set a seed to avoid variations due to partition creation.

        # Set parameters
        self.num_views = 32
        self.num_det_rows = 40
        self.num_det_channels = 128
        self.sharpness = 0.0

        # Initialize sinogram
        self.sinogram_shape = (self.num_views, self.num_det_rows, self.num_det_channels)
        self.angles = None

    def tearDown(self):
        """Clean up after each test method."""
        pass

    def test_qggmrf_denoiser(self):
        """
        Verify that the denoiser works.
        """
        for image_shape in [(1, 100, 100), (100, 100, 100)]:

            max_iterations = 20
            stop_threshold_change_pct = 0.1
            sigma_noise = 0.1
            sharpness = 0.0

            # Get some noisy data
            phantom = np.zeros(image_shape)
            phantom[:, image_shape[1]//3:2*image_shape[1]//3, image_shape[2]//3:2*image_shape[2]//3] = 1
            phantom_noisy = phantom + sigma_noise * np.random.randn(*image_shape)

            denoiser = mj.QGGMRFDenoiser(phantom.shape)
            denoiser.set_params(sharpness=sharpness)
            sigma_noise = 0.1
            phantom_denoised, recon_dict = denoiser.denoise(phantom_noisy, sigma_noise,
                                                            max_iterations=max_iterations,
                                                            stop_threshold_change_pct=stop_threshold_change_pct)
            nrmse = np.linalg.norm(phantom_denoised - phantom) / np.linalg.norm(phantom)
            tolerance = 0.2
            self.assertTrue(nrmse < tolerance)

    def test_median_filter_3d(self):
        a = np.arange(3 * 3 * 3).reshape((3, 3, 3))

        b = mj.median_filter3d(a, return_min_max=True)
        b0 = np.array([[[3., 3., 4.],
                        [6., 6., 7.],
                        [6., 7., 8.]],
                       [[10., 11., 11.],
                        [12., 13., 14.],
                        [15., 15., 16.]],
                       [[18., 19., 20.],
                        [19., 20., 20.],
                        [22., 23., 23.]]])
        b1 = np.array([[[0., 0., 1.],
                        [0., 0., 1.],
                        [3., 3., 4.]],

                       [[0., 0., 1.],
                        [0., 0., 1.],
                        [3., 3., 4.]],

                       [[9., 9., 10.],
                        [9., 9., 10.],
                        [12., 12., 13.]]])
        b2 = np.array([[[13., 14., 14.],
                        [16., 17., 17.],
                        [16., 17., 17.]],

                       [[22., 23., 23.],
                        [25., 26., 26.],
                        [25., 26., 26.]],

                       [[22., 23., 23.],
                        [25., 26., 26.],
                        [25., 26., 26.]]])
        assert np.allclose(b[0], b0)
        assert np.allclose(b[1], b1)
        assert np.allclose(b[2], b2)


if __name__ == '__main__':
    unittest.main()



import jax.numpy as jnp


def generate_direct_recon_filter(num_channels, filter_name="ramp"):
    """
    Creates the specified space domain filter of size (2*num_channels - 1).

    Currently supported filters include: \"ramp\", which corresponds to a ramp in frequency domain.

    Args:
        num_channels (int): Number of detector channels in the sinogram.
        filter_name (string, optional): Name of the filter to be generated. Defaults to "ramp."

    Returns:
        filter (jnp): The computed filter (filter.size = 2*num_channels + 1).
    """

    # If you want to add a new filter, place its name into supported_filters, and ...
    # ... create a new if statement with the filter math.
    # TODO:  Anyone who adds a second filter will need to address how to document the set of available filters
    # in a way that is easy to maintain and appears correctly on readthedocs.  Also, any new filters will need
    # to have the proper scaling.
    supported_filters = ["ramp"]

    # Raise error if filter is not supported.
    if filter_name not in supported_filters:
        raise ValueError(f"Unsupported filter. Supported filters are: {', '.join(supported_filters)}.")

    n = jnp.arange(-num_channels + 1, num_channels)  # ex: num_channels = 3, -> n = [-2, -1, 0, 1, 2]

    recon_filter = 0
    if filter_name == "ramp":
        recon_filter = (1 / 2) * jnp.sinc(n) - (1 / 4) * (jnp.sinc(n / 2)) ** 2

    return recon_filter





"""
Python script to test hdf5 export/import routines.
"""
import numpy as np
import time
import jax.numpy as jnp
import mbirjax as mj

"""**Set the geometry parameters**"""

# Choose the geometry type
geometry_type = 'cone'  # 'cone' or 'parallel'

# Set parameters for the problem size - you can vary these, but if you make num_det_rows very small relative to
# channels, then the generated phantom may not have an interior.
num_views = 128
num_det_rows = 128
num_det_channels = 128

# For cone beam geometry, we need to describe the distances source to detector and source to rotation axis.
# np.Inf is an allowable value, in which case this is essentially parallel beam.
source_detector_dist = 4 * num_det_channels
source_iso_dist = source_detector_dist

# Set parameters for viewing angle.
start_angle = -np.pi
end_angle = np.pi

"""**Data generation:** For demo purposes, we create a phantom and then project it to create a sinogram.

Note:  the sliders on the viewer won't work in notebook form.  For that you'll need to run the python code with an interactive matplotlib backend, typcially using the command line or a development environment like Spyder or Pycharm to invoke python.  

"""

# Initialize sinogram
sinogram_shape = (num_views, num_det_rows, num_det_channels)
angles = jnp.linspace(start_angle, end_angle, num_views, endpoint=False)

if geometry_type == 'cone':
    ct_model_for_generation = mj.ConeBeamModel(sinogram_shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
elif geometry_type == 'parallel':
    ct_model_for_generation = mj.ParallelBeamModel(sinogram_shape, angles)
else:
    raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

# Generate 3D Shepp Logan phantom
print('Creating phantom')
phantom_shape = ct_model_for_generation.get_params('recon_shape')
phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)

# Generate synthetic sinogram data
print('Creating sinogram')
sinogram = ct_model_for_generation.forward_project(phantom)
sinogram = np.asarray(sinogram)

# View sinogram
title = 'Original sinogram \nUse the sliders to change the view or adjust the intensity range.'
mj.slice_viewer(sinogram, slice_axis=0, title=title, slice_label='View')

"""**Initialize for the reconstruction**"""

# ####################
# Initialize the model for reconstruction.
if geometry_type == 'cone':
    ct_model_for_recon = mj.ConeBeamModel(sinogram_shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
elif geometry_type == 'parallel':
    ct_model_for_recon = mj.ParallelBeamModel(sinogram_shape, angles)
else:
    raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

# Print out model parameters
ct_model_for_recon.print_params()

"""**Do the reconstruction and display the results.**"""

# ##########################
# Perform FBP/FDK reconstruction
if geometry_type == 'cone':
    print("Starting FDK recon")
    time0 = time.time()
    recon = ct_model_for_recon.fdk_recon(sinogram, filter_name="ramp")
else:
    print("Starting FBP recon")
    time0 = time.time()
    recon = ct_model_for_recon.fbp_recon(sinogram, filter_name="ramp")

recon.block_until_ready()
elapsed = time.time() - time0
# ##########################

# ##########################
# Test HDF5 import/export functions
mj.export_recon_to_hdf5(filename='../data/recon.h5', recon=recon, recon_description="Test FDK reconstruction")
data = mj.import_recon_from_hdf5('./data/recon.h5')
recon2 = data['recon']
print(data['recon'].shape)
print(data['recon_description'])
print(data['delta_pixel_image'])
print(data['alu_description'])

# Display results
#mj.slice_viewer(recon, recon2, title="Original Recon (left) vs HDF5 Recon (right)")

mj.HDF5Browser()


import numpy as np
import time
import pprint
import jax.numpy as jnp
import mbirjax as mj

if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune the parallel beam mbirjax code.
    """
    # Choose the geometry type
    geometry_type = 'cone'  # 'cone' or 'parallel'

    print('Using {} geometry.'.format(geometry_type))

    # Set parameters
    num_views = 64
    num_det_rows = 64
    num_det_channels = 64
    sharpness = 0.0
    
    # These can be adjusted to describe the geometry in the cone beam case.
    # np.Inf is an allowable value, in which case this is essentially parallel beam
    source_detector_dist = 4 * num_det_channels
    source_iso_dist = source_detector_dist

    if geometry_type == 'cone':
        detector_cone_angle = 2 * np.arctan2(num_det_channels / 2, source_detector_dist)
    else:
        detector_cone_angle = 0
    start_angle = -(np.pi + detector_cone_angle) * (1/2)
    end_angle = (np.pi + detector_cone_angle) * (1/2)

    # Initialize sinogram
    sinogram_shape = (num_views, num_det_rows, num_det_channels)
    angles = jnp.linspace(start_angle, end_angle, num_views, endpoint=False)

    # Set up the model
    if geometry_type == 'cone':
        ct_model = mj.ConeBeamModel(sinogram_shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
    elif geometry_type == 'parallel':
        ct_model = mj.ParallelBeamModel(sinogram_shape, angles)
    else:
        raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

    # Generate 3D Shepp Logan phantom
    print('Creating phantom')
    recon_shape = ct_model.get_params('recon_shape')
    phantom = np.zeros(recon_shape)
    phantom[16:48, 16:48, 16:48] = 1

    # Generate synthetic sinogram data
    print('Creating sinogram')
    sinogram = ct_model.forward_project(phantom)

    # View sinogram
    mj.slice_viewer(sinogram, title='Original sinogram', slice_axis=0, slice_label='View')

    # Generate weights array - for an initial reconstruction, use weights = None, then modify as desired.
    weights = None
    # weights = mj.gen_weights(sinogram / sinogram.max(), weight_type='transmission_root')

    # Set reconstruction parameter values
    ct_model.set_params(sharpness=sharpness, verbose=1)

    # Print out model parameters
    ct_model.print_params()

    # ##########################
    # Perform VCD reconstruction
    time0 = time.time()
    recon, recon_dict = ct_model.recon(sinogram, weights=weights, max_iterations=15)
    recon_params = recon_dict['recon_params']

    recon.block_until_ready()
    elapsed = time.time() - time0
    print('Elapsed time for recon is {:.3f} seconds'.format(elapsed))
    # ##########################

    # Print out parameters used in recon
    pprint.pprint(recon_params)

    max_diff = np.amax(np.abs(phantom - recon))
    print('Geometry = {}'.format(geometry_type))
    nrmse = np.linalg.norm(recon - phantom) / np.linalg.norm(phantom)
    pct_95 = np.percentile(np.abs(recon - phantom), 95)
    print('NRMSE between recon and phantom = {}'.format(nrmse))
    print('Maximum pixel difference between phantom and recon = {}'.format(max_diff))
    print('95% of recon pixels are within {} of phantom'.format(pct_95))

    # Display results
    mj.slice_viewer(phantom, recon, title='Phantom (left) vs VCD Recon (right)')




import numpy as np
import time
import jax.numpy as jnp
import mbirjax as mj

if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune vcd reconstruction
    """
    geometry_type = 'cone'  # 'cone' or 'parallel'

    # Set parameters
    num_views = 64
    num_det_rows = 64
    num_det_channels = 128

    sharpness = 0.0  # This can be used to adjust the regularization
    positivity_flag = True  # Set True to enforce positivity on the recon values

    # These can be adjusted to describe the geometry in the cone beam case.
    # np.Inf is an allowable value, in which case this is essentially parallel beam
    source_detector_dist = 4 * num_det_channels
    source_iso_dist = source_detector_dist

    if geometry_type == 'cone':
        detector_cone_angle = 2 * np.arctan2(num_det_channels / 2, source_detector_dist)
    else:
        detector_cone_angle = 0
    start_angle = -(np.pi + detector_cone_angle) * (1/2)
    end_angle = (np.pi + detector_cone_angle) * (1/2)

    # Initialize sinogram
    sinogram_shape = (num_views, num_det_rows, num_det_channels)
    angles = jnp.linspace(start_angle, end_angle, num_views, endpoint=False)

    # Set up the model
    if geometry_type == 'cone':
        ct_model = mj.ConeBeamModel(sinogram_shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
    elif geometry_type == 'parallel':
        ct_model = mj.ParallelBeamModel(sinogram_shape, angles)
    else:
        raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

    # Here are other things you might want to do

    # Change the recon shape, which has the form (rows, columns, slices)
    # All else being equal, a smaller recon shape will result in a smaller projection on the detector.
    change_recon_shape = False
    if change_recon_shape:
        recon_shape = ct_model.get_params('recon_shape')
        recon_shape = tuple(dim // 2 for dim in recon_shape)  # Or set to whatever you like
        ct_model.set_params(recon_shape=recon_shape)

    # Change the voxel side length (voxels are cubes)
    # The default detector side length is 1.0 (arbitrary units), and delta_voxel has the same units
    # If you change the voxel size, you might want to view the sinogram to see the results.
    # All else being equal, a smaller voxel size will result in a smaller projection on the detector.
    change_voxel_pitch = False
    if change_voxel_pitch:
        ct_model.set_params(delta_voxel=3.0)

    # Change the position of the detector relative to the object
    change_detector_offset = False
    if change_detector_offset:
        ct_model.set_params(det_channel_offset=2.5)  # Positive moves the projected image to the right.
        ct_model.set_params(det_row_offset=5)  # Positive moves the projected image down.

    # Change the granularity as a function of iteration.
    change_partition_sequence = False
    if change_partition_sequence:
        granularity = [1, 4, 8, 64, 256]  # Small granularity means many voxels are updated simultaneously.
        partition_sequence = [0, 1, 2, 3, 4, 2, 3, 2, 3, 3, 3, 3, 3, 3]  # Each entry is an index into granularity.

    # Generate 3D Shepp Logan phantom
    print('Creating phantom')
    phantom_shape = ct_model.get_params('recon_shape')
    phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)
    mj.slice_viewer(phantom, phantom.transpose((0, 2, 1)),
                         title='Phantom\nLeft: single phantom slice (axial)    Right: single phantom row (coronal)',
                         slice_label='Phantom slice', slice_label2='Phantom row', slice_axis2=0)
    # Generate synthetic sinogram data
    print('Creating sinogram: {} geometry'.format(geometry_type))
    sinogram = ct_model.forward_project(phantom)
    # del phantom  # For large recons, you can delete the phantom to preserve memory.

    # View sinogram
    title = 'Original sinogram ({} geometry, {} views)'.format(geometry_type, num_views)
    mj.slice_viewer(sinogram, title=title, slice_label='View', slice_axis=0)

    # Generate weights array
    weights = mj.gen_weights(sinogram / sinogram.max(), weight_type='transmission_root')

    # Set reconstruction parameter values
    ct_model.set_params(sharpness=sharpness, verbose=1)
    ct_model.set_params(positivity_flag=positivity_flag)

    # Print out model parameters
    ct_model.print_params()

    # ##########################
    # Perform VCD reconstruction
    print('Starting recon')
    time0 = time.time()
    recon, recon_dict = ct_model.recon(sinogram, weights=weights)
    recon_params = recon_dict['recon_params']

    recon.block_until_ready()
    elapsed = time.time() - time0
    print('Elapsed time for recon is {:.3f} seconds'.format(elapsed))
    # ##########################

    # Print the results to screen (and file if desired)
    print_summary_to_file = False
    filename = 'recon timing.txt'

    def print_summary(file=None):
        print('\n-------------------------', file=file)
        print('Current stats:', file=file)
        print('Geometry type: {}'.format(geometry_type), file=file)
        print('Sinogram shape (views, rows, channels) = {}'.format(sinogram.shape), file=file)
        print('Recon shape (rows, columns, slices) = {}'.format(recon.shape), file=file)
        print('Sigma_y = {}'.format(recon_params.regularization_params['sigma_y']))
        print('Sigma_x = {}'.format(recon_params.regularization_params['sigma_x']))
        print('Number of iterations = {}'.format(recon_params.num_iterations))
        print('Granularity = {}'.format(recon_params['granularity']))
        print('Partition sequence = {}'.format(recon_params['partition_sequence']))
        print('Final RMSE = {:.3f}'.format(recon_params['fm_rmse'][-1]))
        print('Final prior loss = {:.3f}'.format(recon_params['prior_loss'][-1]))
        print('Elapsed time for recon is {:.3f} seconds'.format(elapsed), file=file)
        mj.get_memory_stats(print_results=True, file=file)
        print('-------------------------', file=file)

    print_summary()

    if print_summary_to_file:
        with open(filename, 'a') as f:
            print_summary(f)

    # Display results
    title = 'VCD recon ({} geometry, {} views)\nLeft: single recon slice (axial)    Right: single recon row (coronal)'.format(geometry_type, num_views)
    mj.slice_viewer(recon, recon.transpose((0, 2, 1)), title=title,
                         slice_label='Recon slice', slice_label2='Recon row', slice_axis2=0)




from dataclasses import dataclass
from typing import Any, Literal
import re
from pathlib import Path
import copy

FILE_FORMAT_NUMBER = 1.0  # The format number should be changed if the file format changes.

# Update to include new geometries that should be included in the tests suite
_geometry_types_for_tests = ['parallel', 'cone', 'translation']

# The order and content of these dictionaries must match the headings and list of dicts below
# The second entry in each case indicates if changing that parameter should trigger a recompile


# Dataclass for parameter storage
@dataclass
class Param:
    val: Any
    recompile_flag: bool = True

    def __repr__(self):
        return f"Param(val={self.val}, recompile_flag={self.recompile_flag})"


# The order and content of these dictionaries must match the headings and list of dicts below
_forward_model_defaults_dict = {
    'geometry_type': Param(None, False),  # The geometry type should never change during a recon.
    'file_format': Param(FILE_FORMAT_NUMBER, False),
    'sinogram_shape': Param(None, True),
    'delta_det_channel': Param(1.0, True),
    'delta_det_row': Param(1.0, True),
    'det_row_offset': Param(0.0, True),
    'det_channel_offset': Param(0.0, True),
    'sigma_y': Param(1.0, False),
}

_recon_model_defaults_dict = {
    'recon_shape': Param(None, True),
    'delta_voxel': Param(None, True),
    'sigma_x': Param(1.0, False),
    'sigma_prox': Param(1.0, False),
    'p': Param(2.0, False),
    'q': Param(1.2, False),
    'T': Param(1.0, False),
    'qggmrf_nbr_wts': Param([1.0, 1.0, 1.0], False),  # Order is row_nbr_wt, col_nbr_wt, slice_nbr_wt
}

_reconstruction_defaults_dict = {
    'auto_regularize_flag': Param(True, False),
    'positivity_flag': Param(False, False),
    'snr_db': Param(30.0, False),
    'sharpness': Param(1.0, False),
    'granularity': Param([1, 2, 4, 8, 16, 32, 64, 128, 256], False),
    'partition_sequence': Param([0, 2, 4, 6, 7], False),
    'verbose': Param(1, False),
    'use_gpu': Param('automatic', True),  # Possible values are 'automatic', 'full', 'sinograms', 'projections', 'none'
}

# These headings should match the dictionaries
headings = ['Forward model parameters', 'Recon parameters', 'Reconstruction parameters']

dicts = [_forward_model_defaults_dict,
         _recon_model_defaults_dict,
         _reconstruction_defaults_dict]

recon_defaults_dict = dict()
for d in dicts:
    recon_defaults_dict = {**recon_defaults_dict, **d}

all_param_keys = list(recon_defaults_dict.keys())


# Utility function to update ParamNames = Literal[...] in parameter_handler.py
def update_param_literal(verify_match_and_exit=False):
    """
    Update the ParamNames = Literal[...] line in parameter_handler.py with the current all_param_keys list.

    - Supports multi-line definitions.
    - If the new and existing definitions match, exits without changes.
    - Otherwise, displays both versions and prompts for confirmation.

    Args:
        verify_match_and_exit (bool): If False, then make the update if needed.  If True, then return status only.

    Returns:
        True if the existing Literal matches the new, otherwise False.
    """
    param_file = Path(__file__).parent / "parameter_handler.py"

    print('Checking for updates to parameter names in ParameterHandler...')

    try:
        content = param_file.read_text()
    except FileNotFoundError:
        print(f"❌ Could not find {param_file}")
        return False

    key_groups = []
    group_size = 4
    for i in range(0, len(all_param_keys), group_size):
        key_groups.append(all_param_keys[i:i+group_size])
    new_param_name = ""
    for group in key_groups:
        new_param_name += "    "
        new_param_name += ", ".join(f"'{key}'" for key in group)
        new_param_name += ",\n"
    new_param_name = f"ParamNames = Literal[\n{new_param_name}]"

    param_name_pattern = re.compile(r"ParamNames\s*=\s*Literal\[[^\]]*?\]", re.DOTALL)
    match_for_name = param_name_pattern.search(content)

    if not match_for_name:
        print("❌ ParamNames = Literal[...] not found in file.")
        return False

    old_param_name = match_for_name.group(0)

    param_names_match = old_param_name.strip() == new_param_name.strip()

    if param_names_match:
        print("✅ ParamNames are already up to date.")
        return True

    if verify_match_and_exit:
        return False

    print('Proposed changes shown below:  additions in green, subtractions in red.')
    if not param_names_match:
        highlight_differences(old_param_name, new_param_name)

    confirm = input("Replace old entries with new ones? (y/n): ").strip().lower()
    if confirm == 'y':
        content = content[:match_for_name.start()] + new_param_name + content[match_for_name.end():]
        param_file.write_text(content)
        print(f"✅ Updated ParamNames in {param_file}")
        return True
    else:
        print("❌ Update cancelled.")
        return False


def get_default_params():
    return copy.deepcopy(recon_defaults_dict)

import difflib
def highlight_differences(string1, string2):
    differ = difflib.Differ()
    diff = differ.compare(string1, string2)
    highlighted_diff = []
    for line in diff:
        if line.startswith('- '): # Highlight string1 differences
            highlighted_diff.append('\033[91m' + line[-1] + '\033[0m')
        elif line.startswith('+ '): # Highlight string2 differences
            highlighted_diff.append('\033[92m' + line[-1] + '\033[0m')
        else:
            highlighted_diff.append(line[-1])
    print(''.join(highlighted_diff))
    # Example usage string1 = "Hello world" string2 = "Hello Python" highlighted = highlight_differences(string1, string2) print(highlighted)



# Script utility to update ParamNames Literal in parameter_handler.py
if __name__ == "__main__":
    update_param_literal()



import numpy as np
import pprint
import jax.numpy as jnp
import mbirjax as mj

if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune the parallel and cone beam mbirjax code.
    """
    # Choose the geometry type
    geometry_type = 'cone'  # 'cone' or 'parallel'

    print('Using {} geometry.'.format(geometry_type))

    # Set parameters
    num_views = 64
    num_det_rows = 40
    num_det_channels = 128
    sharpness = 0.0
    
    # These can be adjusted to describe the geometry in the cone beam case.
    source_detector_dist = 4 * num_det_channels
    source_iso_dist = source_detector_dist

    if geometry_type == 'cone':
        detector_cone_angle = 2 * np.arctan2(num_det_channels / 2, source_detector_dist)
    else:
        detector_cone_angle = 0
    start_angle = -(np.pi + detector_cone_angle) * (1/2)
    end_angle = (np.pi + detector_cone_angle) * (1/2)

    # Initialize sinogram
    sinogram_shape = (num_views, num_det_rows, num_det_channels)
    angles = jnp.linspace(start_angle, end_angle, num_views, endpoint=False)

    # Set up the model
    if geometry_type == 'cone':
        ct_model = mj.ConeBeamModel(sinogram_shape, angles, source_detector_dist=source_detector_dist, source_iso_dist=source_iso_dist)
    elif geometry_type == 'parallel':
        ct_model = mj.ParallelBeamModel(sinogram_shape, angles)
    else:
        raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

    # Generate 3D Shepp Logan phantom
    print('Creating phantom')
    recon_shape = ct_model.get_params("recon_shape")
    phantom = mj.generate_3d_shepp_logan_low_dynamic_range(recon_shape)

    # Generate synthetic sinogram data
    print('Creating sinogram')
    sinogram = ct_model.forward_project(phantom)

    # View sinogram
    mj.slice_viewer(sinogram, title='Original sinogram', slice_axis=0, slice_label='View')

    # Generate weights array - for an initial reconstruction, use weights = None, then modify as desired.
    weights = None
    # weights = mj.gen_weights(sinogram / sinogram.max(), weight_type='transmission_root')

    # Set reconstruction parameter values
    ct_model.set_params(sharpness=sharpness, verbose=1)

    # Print out model parameters
    ct_model.print_params()

    # Perform MBIR reconstruction
    recon, recon_params = ct_model.recon(sinogram, weights=weights)

    # Print out parameters used in recon
    pprint.pprint(recon_params['recon_params'])

    # Use export/import functions to write out recon to disk
    import os
    os.makedirs('output', exist_ok=True)
    mj.export_recon_hdf5('output/recon.h5', recon)
    recon, _ = mj.import_recon_hdf5('output/recon.h5')

    # Display results
    mj.slice_viewer(phantom, recon, title='Phantom (left) vs MBIR Recon (right)')



import io
import datetime
from typing import Literal, Union, overload, Any
import numpy as np
import jax
import jax.numpy as jnp
import jax.lax as lax
import mbirjax as mj
from mbirjax import TomographyModel

QGGMRFDenoiserParamNames = mj.ParamNames | Literal['sigma_noise']


class QGGMRFDenoiser(TomographyModel):
    """
    The QGGMRFDenoiser uses the MBIRJAX recon framework to implement a qggmrf proximal map denoiser.
    The primary interface is through :meth:`denoise`.
    """

    def __init__(self, image_shape):

        view_params_name = 'None'
        if len(image_shape) != 3:
            raise ValueError('image_shape must be 3-dimensional. Got image_shape={}. To denoise a 2D image, use shape (1, m, n).'.format(image_shape))
        super().__init__(image_shape, view_params_name=view_params_name, sigma_noise=None)
        self.use_ror_mask = False
        self.set_params(sharpness=0)  # The default sharpness level is 0 for the denoiser.

        self.set_params(granularity=[16], partition_sequence=[0])  # For qggmrf denoising, we can fix a partition
        try:
            jax.devices('gpu')
            self.set_params(use_gpu='full')  # If the denoising problem doesn't fit on the gpu, we should divide it up.
        except RuntimeError:
            self.set_params(use_gpu='none')

    @overload
    def get_params(self, parameter_names: Union[QGGMRFDenoiserParamNames, list[QGGMRFDenoiserParamNames]]) -> Any: ...

    def get_params(self, parameter_names) -> Any:
        return super().get_params(parameter_names)

    def get_magnification(self):
        """
        Return 1 to satisfy the TomographyModel interface.

        Returns:
            (float): magnification
        """
        magnification = 1.0
        return magnification

    def verify_valid_params(self):
        """
        Check that all parameters are compatible for a reconstruction.

        Note:
            Raises ValueError for invalid parameters.
        """
        super().verify_valid_params()
        sinogram_shape = self.get_params('sinogram_shape')

        image_shape = self.get_params('recon_shape')
        if image_shape != sinogram_shape:
            error_message = "image_shape and sinogram_shape must be the same. \n"
            error_message += "Got {} for image_shape and {} for sinogram_shape".format(image_shape, sinogram_shape)
            raise ValueError(error_message)

    def get_geometry_parameters(self):
        raise NotImplementedError('get_geometry_parameters is not implemented for QGGMRFDenoiser.')

    def create_projectors(self):
        pass

    def auto_set_recon_shape(self, sinogram_shape, no_compile=True, no_warning=False):
        """Compute the default recon shape to equal the sinogram shape"""
        self.set_params(no_compile=no_compile, no_warning=no_warning, recon_shape=sinogram_shape)

    def auto_set_sigma_y(self, sinogram, sino_indicator, weights=1):
        sigma_y = self.get_params('sigma_noise')
        self.set_params(no_warning=True, sigma_y=sigma_y, auto_regularize_flag=True)

    def estimate_image_noise_std(self, image):
        """
        Estimate the standard deviation of the reconstruction from the noisy image.

        Args:
            image (jax array or ndarray): 3D array containing noisy image with shape (num_views, num_det_rows, num_det_channels).
        """
        support_indicator = self._get_sino_indicator(image, sigma_noise=0.0)
        sigma_noise = self._get_estimate_of_recon_std(image, support_indicator)
        support_indicator = self._get_sino_indicator(image, sigma_noise=sigma_noise)
        sigma_noise = self._get_estimate_of_recon_std(image, support_indicator)
        return sigma_noise

    def _get_estimate_of_recon_std(self, noisy_image, support_indicator):
        """
        Estimate the standard deviation of the reconstruction from the noisy image.  This is used to scale sigma_prox and
        sigma_x in MBIR reconstruction.

        Args:
            noisy_image (ndarray): 3D jax array containing noisy image with shape (num_views, num_det_rows, num_det_channels).
            support_indicator (ndarray): a binary mask that indicates the region of image support; same shape as noisy_image.
        """

        # # Compute the typical magnitude of a noisy image value
        # typical_image_value = np.average(np.abs(noisy_image), weights=support_indicator)
        # return typical_image_value

        inds = np.where(support_indicator)

        vals = np.stack([noisy_image[inds[0], inds[1], inds[2]],
                        noisy_image[inds[0]-1, inds[1], inds[2]],
                        noisy_image[inds[0], inds[1]-1, inds[2]],
                        noisy_image[inds[0], inds[1], inds[2]-1]], axis=0)
        std = np.std(vals, axis=0)
        recon_std = np.mean(std)

        return recon_std  # typical_image_value

    def _get_sino_indicator(self, noisy_image, sigma_noise=None):
        """
        Compute a binary mask that indicates the region of noisy_image support.

        Args:
            noisy_image (jax array or ndarray): 3D array containing noisy_image with shape (num_views, num_det_rows, num_det_channels).
            sigma_noise (float, optional): Estimated noise standard deviation in the image.  If None, then this is estimated from the image.

        Returns:
            (ndarray): Weights used in mbircone reconstruction, with the same array shape as ``noisy_image``.
        """
        if sigma_noise is None:
            sigma_noise = self.get_params('sigma_noise')

        percent_noise_floor = 5.0
        # Form indicator by thresholding noisy_image
        threshold = (0.01 * percent_noise_floor) * np.mean(np.fabs(noisy_image)) + sigma_noise
        threshold = min(threshold, np.amax(noisy_image))
        indicator = np.int8(noisy_image >= threshold)
        return indicator

    def recon(self, *args, **kwargs):
        raise NotImplementedError('recon is not implemented for QGGMRFDenoiser.  Use `denoise` instead.')

    def denoise(self, image, sigma_noise=None, use_ror_mask=False, init_image=None, max_iterations=15,
                stop_threshold_change_pct=0.2, first_iteration=0, logfile_path='./logs/recon.log', print_logs=True):
        """
        Compute the MAP denoiser assuming AWGN and the 3D qGGMRF prior.

        With default settings, and with X a clean image and W equal to AWGN of standard deviation sigma_noise,
        the result of :meth:`denoise` applied to X+W is the MAP estimate of the denoised image using the
        qGGMRF prior function.

        The amount of denoising can be changed by changing sigma_noise.  If sigma_noise is None, then sigma_noise
        is estimated from a set of samples from the image.

        Denoising strength can also be adjusted using the parameter `sharpness` (default=0.0).

        Args:
            image (numpy or jax array):  The 3D volume to be denoised.
            sigma_noise (float, optional):  The estimated noise variance in the noisy image.  If None, then the noise level is estimated from the image.
            use_ror_mask (bool, optional): Set true to restrict denoising to an inscribed circle in the image.  Defaults to False.
            init_image (numpy or jax array, optional):  An initial image for the minimization.  Defaults to image.
            max_iterations (int, optional): maximum number of iterations of the VCD algorithm to perform.
            stop_threshold_change_pct (float, optional): Stop reconstruction when 100 * ||delta_recon||_1 / ||recon||_1 change from one iteration to the next is below stop_threshold_change_pct.  Defaults to 0.2.  Set this to 0 to guarantee exactly max_iterations.
            first_iteration (int, optional): Set this to be the number of iterations previously completed when restarting a recon using init_recon.  This defines the first index in the partition sequence.  Defaults to 0.
            logfile_path (str, optional): Path to the output log file.  Defaults to './logs/recon.log'.
            print_logs (bool, optional): If true then print logs to console.  Defaults to True.

        Returns:
            tuple: (denoised_image, denoiser_dict)
                - denoised_image (jax array): A denoised image of the same shape as image
                - denoiser_dict (dict): A dict obtained from :meth:`get_recon_dict` with entries
                    * 'recon_params'
                    * 'notes'
                    * 'recon_logs'
                    * 'model_params'

        Example:
            >>> denoiser = mj.QGGMRFDenoiser(noisy_image.shape)
            >>> denoiser.set_params(sharpness=0.5)  # Increase sharpness a little over the default of 0.0
            >>> denoised_image, denoised_dict = denoiser.denoise(noisy_image)  # Estimate the noise level from the image
            >>> mj.slice_viewer(noisy_image, denoised_image, data_dicts=[None, denoised_dict], title='Noisy and denoised images')

        See Also
        --------
        TomographyModel : The base class from which this class inherits.
        """
        self.use_ror_mask = use_ror_mask
        if sigma_noise is None:
            sigma_noise = self.estimate_image_noise_std(image)
        self.set_params(sigma_noise=sigma_noise)
        if first_iteration == 0 or self.logger is None:
            self.setup_logger(logfile_path=logfile_path, print_logs=print_logs)

        self.logger.info('Initializing QGGMRFDenoiser')
        regularization_params = self.auto_set_regularization_params(image)

        # Generate set of voxel partitions
        image_shape, granularity = self.get_params(['recon_shape', 'granularity'])
        partition_sequence = self.get_params('partition_sequence')
        partition_index = partition_sequence[0]
        partitions = mj.gen_set_of_pixel_partitions(image_shape, [granularity[partition_index]], use_ror_mask=self.use_ror_mask)

        # Generate sequence of partitions to use
        partition = partitions[0]

        if init_image is None:
            init_image = image.copy()

        flat_error_image = (image - init_image).reshape((-1, image.shape[-1]))
        flat_error_image = jax.device_put(flat_error_image, self.worker)
        image_out = init_image

        verbose, sigma_y = self.get_params(['verbose', 'sigma_y'])

        # Initialize the output image
        flat_image = image_out.reshape((-1, image_shape[2]))
        flat_image = jax.device_put(flat_image, self.worker)

        # Create the finer grained image update operators

        fm_constant = 1.0 / (self.get_params('sigma_y') ** 2.0)
        qggmrf_nbr_wts, sigma_x, p, q, T = self.get_params(['qggmrf_nbr_wts', 'sigma_x', 'p', 'q', 'T'])
        b = mj.get_b_from_nbr_wts(qggmrf_nbr_wts)
        qggmrf_params = tuple((b, sigma_x, p, q, T))
        image_shape = self.get_params('recon_shape')

        @jax.jit  # JIT the whole sweep
        def denoise_over_partition(local_flat_image, local_flat_error_image):
            """Run vcd_subset_denoiser over every subset in `partition`
            using lax.fori_loop to keep the loop on-device and JIT-compatible."""
            # Analog of vcd_partition_iterator

            # Bundle *all* mutable state into one carry object
            def body_fn(i, carry):
                body_flat_image, body_flat_error_image, ell1_accum, alpha_accum = carry

                subset = partition[i]  # pick i-th subset

                body_flat_image, body_flat_error_image, ell1_subset, alpha_subset = (
                    vcd_subset_denoiser(body_flat_image, body_flat_error_image,
                                        subset, fm_constant, qggmrf_params, image_shape))

                # update running totals
                ell1_accum = ell1_accum + ell1_subset
                alpha_accum = alpha_accum + alpha_subset

                return (body_flat_image, body_flat_error_image,
                        ell1_accum, alpha_accum)

            # initial carry
            init_carry = (local_flat_image, local_flat_error_image,
                          0.0, 0.0)

            # run 0 … N-1
            local_flat_image, local_flat_error_image, local_ell1, local_alpha = (
                lax.fori_loop(0, partition.shape[0], body_fn, init_carry))
            final_carry = (local_flat_image, local_flat_error_image, local_ell1, local_alpha / partition.shape[0])
            return final_carry  # unpack outside if you like

        # pre-allocate history arrays (static length = max_iters)
        max_iters = max_iterations
        nmae_update_init = jnp.zeros(max_iters)
        alpha_values_init = jnp.zeros(max_iters)

        stop_thresh = stop_threshold_change_pct / 100.0  # scalar threshold

        def log_updates(updates):
            cur_iter, cur_nmae = updates
            iter_output = 'After iteration {} of a max of {}: Pct change={:.4f}'.format(cur_iter + first_iteration, max_iters, 100 * cur_nmae)
            self.logger.info(iter_output)

        @jax.jit
        def run_denoising_loop(local_flat_image, local_flat_error_image):
            """
            Runs the outer optimisation loop with lax.while_loop.
            Returns:
                (local_flat_image, local_flat_error_image,
                nmae_hist, alpha_hist, num_iters)
            """

            # -------- carry = all mutable state --------
            # i                 : iteration counter
            # local_flat_image        : current image
            # local_flat_error_image  : current residual
            # nmae_hist         : history array (filled progressively)
            # alpha_hist        : "
            # nmae_curr         : nmae from *most recent* iteration
            carry0 = (0, local_flat_image, local_flat_error_image,
                      nmae_update_init, alpha_values_init, jnp.inf  # nmae_curr (start high so loop begins)
                      )

            # -------- termination condition --------
            def cond_fn(carry):
                i, *_, nmae_curr = carry
                not_enough_iters = i < max_iters
                change_big_enough = nmae_curr >= stop_thresh
                return jnp.logical_and(not_enough_iters, change_big_enough)

            # -------- body: one outer iteration --------
            def body_fn(carry):
                (i, flat_img, flat_err_img, nmae_body, alpha_body, _) = carry

                # inner loop (already JAX-ified with fori_loop previously)
                flat_img, flat_err_img, ell1_for_part, alpha_val = \
                    denoise_over_partition(flat_img, flat_err_img)

                # --- compute stats ---
                nmae = ell1_for_part / jnp.sum(jnp.abs(flat_img))

                # --- write into the history arrays ---
                nmae_body = nmae_body.at[i].set(nmae)
                alpha_body = alpha_body.at[i].set(alpha_val)

                # Insert call back to print progress
                print_rate = 5
                iter_updates = (i, nmae_body[i])
                _ = lax.cond(
                    (i % print_rate) == 0,
                    lambda c: (
                        # call host_callback then return unchanged carry
                        jax.debug.callback(log_updates, iter_updates)
                    ),
                    lambda c: c,
                    None
                )

                # bump iteration counter
                return i + 1, flat_img, flat_err_img, nmae_body, alpha_body, nmae  # new nmae for cond_fn

            # run the while-loop
            final_carry = lax.while_loop(cond_fn, body_fn, carry0)

            # unpack results
            (num_iters_loop, local_flat_image, local_flat_error_image,
             nmae_hist, alpha_hist, _) = final_carry

            return local_flat_image, local_flat_error_image, nmae_hist, alpha_hist, num_iters_loop

        self.logger.info('Starting VCD iterations')
        if verbose >= 2:
            output = io.StringIO()
            mj.get_memory_stats(file=output)
            self.logger.debug(output.getvalue())
            self.logger.debug('--------')

        # Do the iterations
        (flat_image, flat_error_image,  # full length = max_iters (unused slots stay 0)
         nmae_update, alpha_values, num_iters) = run_denoising_loop(flat_image, flat_error_image)

        fm_rmse = None
        recon_params = (fm_rmse, nmae_update[0:num_iters], alpha_values[0:num_iters])
        stop_threshold_change_pct = [100 * float(val) for val in recon_params[1]]
        alpha_values = [float(val) for val in recon_params[2]]

        prior_loss = None
        recon_param_values = [int(num_iters), granularity, partition_sequence, fm_rmse, prior_loss,
                              regularization_params, stop_threshold_change_pct, alpha_values]
        recon_params = mj.ReconParams(*tuple(recon_param_values))._asdict()

        notes = 'Reconstruction completed: {}\n\n'.format(datetime.datetime.now())
        denoiser_dict = self.get_recon_dict(recon_params, notes=notes)
        denoised_image = self.reshape_recon(flat_image)

        return denoised_image, denoiser_dict


def vcd_subset_denoiser(flat_image, flat_error_image, pixel_indices,
                        fm_constant, qggmrf_params, image_shape):
    # This is the analog of vcd_subset_updater
    # qGGMRF prior - compute the qggmrf gradient and hessian at each pixel in the index set.
    prior_grad, prior_hess = (
        mj.qggmrf_gradient_and_hessian_at_indices(flat_image, image_shape, pixel_indices,
                                                  qggmrf_params))

    # Back project to get the gradient - the forward Hessian is all 1s for the qggmrf proximal map
    cur_error_image = flat_error_image[pixel_indices]
    forward_grad = - fm_constant * cur_error_image
    forward_hess = 1

    # Compute update vector update direction in recon domain
    delta_recon_at_indices = - ((forward_grad + prior_grad) / (forward_hess + prior_hess))

    # Compute delta^T \nabla Q(x_hat; x'=x_hat) for use in finding alpha
    prior_linear = jnp.sum(prior_grad * delta_recon_at_indices)

    # Estimated upper bound for hessian
    prior_overrelaxation_factor = 1.1
    prior_quadratic_approx = ((1 / prior_overrelaxation_factor) *
                              jnp.sum(prior_hess * delta_recon_at_indices ** 2))

    # Compute update direction in sinogram domain
    delta_sinogram = delta_recon_at_indices
    forward_linear = fm_constant * jnp.tensordot(cur_error_image, delta_sinogram, axes=2)
    forward_quadratic = fm_constant * jnp.tensordot(delta_sinogram, delta_sinogram, axes=2)

    # Compute optimal update step
    alpha_numerator = forward_linear - prior_linear
    alpha_denominator = forward_quadratic + prior_quadratic_approx + jnp.finfo(jnp.float32).eps
    alpha = alpha_numerator / alpha_denominator
    max_alpha = 1.5
    alpha = jnp.clip(alpha, jnp.finfo(jnp.float32).eps, max_alpha)

    delta_recon_at_indices = alpha * delta_recon_at_indices
    flat_image = flat_image.at[pixel_indices].add(delta_recon_at_indices)

    # Update sinogram and loss
    cur_error_image = cur_error_image - alpha * delta_sinogram
    flat_error_image = flat_error_image.at[pixel_indices].set(cur_error_image)
    ell1_for_subset = jnp.sum(jnp.abs(delta_recon_at_indices))
    alpha_for_subset = alpha

    return flat_image, flat_error_image, ell1_for_subset, alpha_for_subset


def median_filter3d(x, max_block_gb=4.0, return_min_max=False) -> Union[jnp.ndarray | tuple]:
    """
    Apply a 27‑point (3x3x3) median filter to a 3‑D JAX array using replicated
    (edge) boundary conditions.  Optionally return the min and max in each 27 point neighborhood.

    The volume is processed in d0‑blocks so that the kernel can be
    `jax.jit`‑compiled while limiting peak device memory. Each block is padded with
    a one‑voxel halo; halos duplicate the nearest edge voxel so that the result
    matches NumPy’s `"edge"` mode.

    Args:
        x (jax array or ndarray): Input array. Any numeric dtype supported by JAX is allowed.
        max_block_gb (float. optional): A rough upper bound on the amount of memory in GB to use for the filtering.  Defaults to 4.0.
        return_min_max (bool, optional): If true, the output is a tuple of median, min, max.

    Returns:

        jax.numpy.ndarray or tuple: An array (or tuple of 3 arrays) of the same shape and dtype as *x* containing the median‑filtered result.

    Notes
    -----
    * The function automatically splits the 0‑dimension into blocks so that at
      most roughly ``max_block_gb`` of temporary data are materialised.
      If the array is large and the 0 dimension is small relative to another dimension, it may be more memory efficient
      to apply jnp.swapaxes(x, 0, long_dim) before applying median_filter3d, although swapaxes will make a copy of x.
    * Within each block the filter is computed by rolling the data in all 26
      neighbour directions, stacking the 27 volumes, and taking
      :func:`jnp.median` along the new axis.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import mbirjax as mj
    >>> vol = jnp.arange(27.).reshape(3, 3, 3)
    >>> mj.median_filter3d(vol)
    Array([[[3., 3., 4.],
            [6., 6., 7.],
            [6., 7., 8.]],
           ...
           dtype=float32)
    """
    d0, d1, d2 = x.shape
    x_gb = x.size * 4 / (1024**3)
    num_blocks = np.ceil(27 * x_gb / max_block_gb).astype(int)
    block_size = max(d0 // num_blocks, 1)
    # 1) Pad d2 in every dim by 1 for the edge‐replicated halo
    xp = jnp.pad(x, pad_width=1, mode='edge')      # (d0+2, d1+2, d2+2)

    # 2) Pad d0 *further* up to a multiple of block_size
    n_blocks = (d0 + block_size - 1) // block_size
    padded_Z = n_blocks * block_size
    # pad only at the *end* in d0 so that we can dynamic‐slice fixed size blocks
    pad_extra = padded_Z - d0
    xp = jnp.pad(xp,
                 pad_width=((0, pad_extra), (0,0), (0,0)),
                 mode='edge')                        # (padded_Z+2, d1+2, d2+2)

    # fixed slice size for every block
    slice_sz = (block_size + 2,  d1 + 2,  d2 + 2)

    def filter_one_block(i):
        # compute the dynamic start index in the padded xp
        z0 = i * block_size
        start = (z0, 0, 0)
        # grab a fixed‐shape window [block_size+2, d1+2, d2+2]
        block = lax.dynamic_slice(xp, start, slice_sz)

        # replicate the 27‐roll → stack → median recipe on this small block
        patches = [
          jnp.roll(block, shift=(dz,dy,dx), axis=(0,1,2))
          for dz in (-1,0,1) for dy in (-1,0,1) for dx in (-1,0,1)
        ]
        stacked  = jnp.stack(patches, axis=0)      # (27, blkZ+2, d1+2, d2+2)
        filtered = jnp.median(stacked, axis=0)    # (blkZ+2, d1+2, d2+2)
        if return_min_max:
            min_filtered = jnp.min(stacked, axis=0)
            max_filtered = jnp.max(stacked, axis=0)
            filtered = jnp.stack([filtered, min_filtered, max_filtered], axis=3)

        # strip off the 1‐voxel halo in all dims → (blkZ, d1, d2)
        return filtered[1:-1, 1:-1, 1:-1]

    # 3) Map over blocks
    blocks = lax.map(filter_one_block, jnp.arange(n_blocks))  # (n_blocks, blkZ, d1, d2)

    # 4) Stitch & then crop back to original d0
    out = jnp.concatenate(blocks, axis=0)  # (padded_Z, d1, d2)
    out = out[:d0, :, :]                   # (d0, d1, d2)
    if return_min_max:
        out = (out[..., 0], out[..., 1], out[..., 2])
    return out



from .projectors import *
from .parameter_handler import *
from .tomography_model import *
from .qggmrf import *
from .parallel_beam import *
from .cone_beam import *
from .denoising import *
from .vcd_utils import *
from .memory_stats import *
from .utilities import *
from .viewer import *
from . import preprocess
from .translation_model import *


import jax
import jax.numpy as jnp
from functools import partial
from collections import namedtuple
from mbirjax import TomographyModel, ParameterHandler


class TemplateModel(TomographyModel):
    """
    This is a template for a class designed to handle a particular geometry that extends the :ref:`TomographyModelDocs`.
    This class inherits all methods and properties from the :ref:`TomographyModelDocs` and may override some
    to suit the specific geometry.

    Parameters not included in the constructor can be set using the set_params method of :ref:`TomographyModelDocs`.
    Refer to :ref:`TomographyModelDocs` documentation for a detailed list of possible parameters.

    Items to change for a particular geometry are highlighted with TODO.
    It is not recommended to use **kwargs in the constructor since doing so complicates checking for valid parameter
    names in set_params.

    Args:
        sinogram_shape (tuple):
            Shape of the sinogram as a tuple in the form `(views, rows, channels)`, where 'views' is the number of
            different projection angles, 'rows' correspond to the number of detector rows, and 'channels' index columns of
            the detector that are assumed to be aligned with the rotation axis.
        param1, param2 (scalars):
            These are view-independent scalar parameters that are required for the geometry and are not already included in the parent class.
        view_dependent_vec1, view_dependent_vec2 (jnp.ndarray):
            These are view-dependent parameter vectors each with length = number of views.
    """

    # TODO: Adjust the signature as needed for a particular geometry and update the docstring to match.
    # Don't include any additional unspecified keyword arguments in the form of **kwargs - use only the parameters that
    # are required for the geometry.  Any changes to existing parameters should be done by the user with set_params,
    # which checks for invalid parameter names.  There is no check for invalid parameter names here because the
    # geometry may need to define new parameters.
    def __init__(self, sinogram_shape, param1, param2, view_dependent_vec1, view_dependent_vec2):
        # Convert the view-dependent vectors to an array
        view_dependent_vecs = [vec.flatten() for vec in [view_dependent_vec1, view_dependent_vec2]]
        try:
            view_params_array = jnp.stack(view_dependent_vecs, axis=1)
        except ValueError as e:
            raise ValueError("Incompatible view dependent vector lengths:  all view-dependent vectors must have the "
                             "same length.")

        super().__init__(sinogram_shape, param1=param1, param2=param2, view_params_array=view_params_array)

    @classmethod
    def from_file(cls, filename):
        """
        Construct a ConeBeamModel from parameters saved using save_params()

        Args:
            filename (str): Name of the file containing parameters to load.

        Returns:
            ConeBeamModel with the specified parameters.
        """
        # Load the parameters and convert view-dependent parameters to use the geometry-specific keywords.
        # TODO: Adjust these to match the signature of __init__
        required_param_names = ['sinogram_shape', 'param1', 'param2']
        required_params, params = ParameterHandler.load_param_dict(filename, required_param_names, values_only=True)

        # TODO: Adjust these to match the signature of __init__
        view_params_array = params['view_params_array']
        required_params['view_dependent_vec1'] = view_params_array[:, 0]
        required_params['view_dependent_vec2'] = view_params_array[:, 1]
        del params['view_params_array']

        new_model = cls(**required_params)
        new_model.set_params(**params)
        return new_model

    def get_magnification(self):
        """
        Compute the scale factor from a voxel at iso (at the origin on the center of rotation) to
        its projection on the detector.  For parallel beam, this is 1, but it may be parameter-dependent
        for other geometries.

        Returns:
            (float): magnification
        """
        # TODO: Adjust as needed for the geometry.
        magnification = 1.0
        return magnification

    def verify_valid_params(self):
        """
        Check that all parameters are compatible for a reconstruction.
        """
        super().verify_valid_params()
        sinogram_shape, view_params_array = self.get_params(['sinogram_shape', 'view_params_array'])

        # TODO: Modify as needed for the geometry.
        if view_params_array.shape[0] != sinogram_shape[0]:
            error_message = "Number view dependent parameter vectors must equal the number of views. \n"
            error_message += "Got {} for length of view-dependent parameters and "
            error_message += "{} for number of views.".format(view_params_array.shape[0], sinogram_shape[0])
            raise ValueError(error_message)

    def get_geometry_parameters(self):
        """
        Required function to get the view independent geometry parameters required for projection.

        Returns:
            namedtuple of any parameters required for back_project_one_view_to_pixel_batch or forward_project_pixel_batch_to_one_view.
            This does not need to include view dependent parameters, or sinogram_shape or recon_shape, which
            are passed in automatically to projector_params.
        """
        # TODO: Include additional names as needed for the projectors.
        # First get the parameters managed by ParameterHandler
        geometry_param_names = ['delta_det_channel', 'det_channel_offset', 'delta_voxel']
        geometry_param_values = self.get_params(geometry_param_names)

        # Then get additional parameters that are calculated separately, such as psf_radius and magnification.
        # geometry_param_names += ['psf_radius']
        # geometry_param_values.append(self.get_psf_radius())
        # geometry_param_names += ['magnification']
        # geometry_param_values.append(self.get_magnifiction())

        # Then create a namedtuple to access parameters by name in a way that can be jit-compiled.
        GeometryParams = namedtuple('GeometryParams', geometry_param_names)
        geometry_params = GeometryParams(*tuple(geometry_param_values))

        return geometry_params

    def auto_set_recon_shape(self, sinogram_shape, no_compile=True, no_warning=False):
        """Compute the default recon size using the internal parameters delta_channel and delta_pixel plus
          the number of channels from the sinogram"""
        # TODO: provide code to implement this
        raise NotImplementedError('auto_set_recon_shape must be implemented by each specific geometry model.')

    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def forward_project_pixel_batch_to_one_view(voxel_values, pixel_indices, single_view_params, projector_params):
        """
        Forward project a set of voxels determined by indices into a single view.

        NOTE: This function must be able to be jit-compiled.

        Args:
            voxel_values (jax array):  2D array of shape (num_indices, num_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for this view.
            projector_params (1D jax array): tuple of (sinogram_shape, recon_shape, get_geometry_params()).

        Returns:
            jax array of shape (num_det_rows, num_det_channels)
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params  # This is the namedtuple from get_geometry_parameters
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        num_recon_rows, num_recon_columns, num_recon_slices = projector_params.recon_shape

        # Returns a single view of the sinogram
        sinogram_view = jnp.zeros((num_det_rows, num_det_channels))

        # TODO:  Provide code to implement forward projection

        return sinogram_view

    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def back_project_one_view_to_pixel_batch(sinogram_view, pixel_indices, single_view_params, projector_params, coeff_power=1):
        """
        Calculate the backprojection value at a specified recon voxel cylinders given a sinogram view and various parameters.

        NOTE: This function must be able to be jit-compiled.

        Args:
            sinogram_view (2D jax array): one view of the sinogram to be back projected
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for the view being back projected.
            projector_params (1D jax array): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing theta 2.

        Returns:
            The voxel values for all slices at the input index (i.e., a voxel cylinder) obtained by backprojecting
            the input sinogram view.
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params  # This is the namedtuple from get_geometry_parameters
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        num_recon_rows, num_recon_columns, num_recon_slices = projector_params.recon_shape

        # Computes the voxel values in all slices corresponding to pixel_index
        num_pixels = pixel_indices.shape[0]
        voxel_values_cylinder = jnp.zeros((num_pixels, num_recon_slices))

        # TODO:  Provide code to implement forward projection

        return voxel_values_cylinder



import os
import numpy as np
import jax.numpy as jnp
import jax
import time
import matplotlib.pyplot as plt
import gc
import mbirjax as mj

if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune the parallel beam projector
    """
    # ##########################
    # Do all the setup
    view_batch_size = 32
    pixel_batch_size = 2048

    # Initialize sinogram
    num_views = 128
    num_det_rows = 20
    num_det_channels = 128
    start_angle = 0
    end_angle = jnp.pi
    sinogram = jnp.zeros((num_views, num_det_rows, num_det_channels))
    angles = jnp.linspace(start_angle, jnp.pi, num_views, endpoint=False)

    # Initialize a random key
    seed_value = np.random.randint(1000000)
    key = jax.random.PRNGKey(seed_value)

    # Set up parallel beam model
    # parallel_model = mj.ParallelBeamModel.from_file('params_parallel.yaml')
    parallel_model = mj.ParallelBeamModel(sinogram.shape, angles)
    # parallel_model.to_file('params_parallel.yaml')

    # Generate phantom
    recon_shape = parallel_model.get_params('recon_shape')
    num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
    phantom = mj.gen_cube_phantom(recon_shape)

    # Generate indices of pixels
    num_subsets = 1
    full_indices = mj.gen_pixel_partition(recon_shape, num_subsets)
    num_subsets = 5
    subset_indices = mj.gen_pixel_partition(recon_shape, num_subsets)

    # Generate sinogram data
    voxel_values = phantom.reshape((-1,) + recon_shape[2:])[full_indices]

    parallel_model.set_params(view_batch_size=view_batch_size, pixel_batch_size=pixel_batch_size)

    print('Starting forward projection')
    sinogram = parallel_model.sparse_forward_project(voxel_values[0], full_indices[0])

    # Determine resulting number of views, slices, and channels and image size
    print('Sinogram shape: {}'.format(sinogram.shape))
    print('Memory stats after forward projection')
    mj.get_memory_stats(print_results=True)

    # Get the vector of indices
    indices = jnp.arange(num_recon_rows * num_recon_cols)
    num_trials = 3
    indices = jnp.mod(np.arange(num_trials, dtype=int).reshape((-1, 1)) + indices.reshape((1, -1)), num_recon_rows * num_recon_cols)

    sinogram = jnp.array(sinogram)
    indices = jnp.array(indices)

    # Run once to finish compiling
    print('Starting back projection')
    bp = parallel_model.sparse_back_project(sinogram, indices[0])
    print('Recon shape: ({}, {}, {})'.format(num_recon_rows, num_recon_cols, num_recon_slices))
    print('Memory stats after back projection')
    mj.get_memory_stats(print_results=True)
    # ##########################
    # Test the adjoint property
    # Get a random 3D phantom to test the adjoint property
    key, subkey = jax.random.split(key)
    x = jax.random.uniform(subkey, shape=bp.shape)
    key, subkey = jax.random.split(key)
    y = jax.random.uniform(subkey, shape=sinogram.shape)

    # Do a forward projection, then a backprojection
    voxel_values = x.reshape((-1, num_recon_slices))[indices[0]]
    Ax = parallel_model.sparse_forward_project(voxel_values, indices[0])
    Aty = parallel_model.sparse_back_project(y, indices[0])

    # Calculate <Aty, x> and <y, Ax>
    Aty_x = jnp.sum(Aty * x)
    y_Ax = jnp.sum(y * Ax)

    print("Adjoint property holds for random x, y <y, Ax> = <Aty, x>: {}".format(np.allclose(Aty_x, y_Ax)))

    # Clean up before further projections
    del Ax, Aty, bp
    del phantom
    del x, y
    gc.collect()

    # ##########################
    # ## Test the hessian against a finite difference approximation ## #
    hessian = parallel_model.compute_hessian_diagonal()

    x = jnp.zeros(recon_shape)
    key, subkey = jax.random.split(key)
    i, j = jax.random.randint(subkey, shape=(2,), minval=0, maxval=num_recon_rows)
    key, subkey = jax.random.split(key)
    k = jax.random.randint(subkey, shape=(), minval=0, maxval=num_recon_slices)

    eps = 0.01
    x = x.at[i, j, k].set(eps)
    voxel_values = x.reshape((-1, num_recon_slices))[indices[0]]
    Ax = parallel_model.sparse_forward_project(voxel_values, indices[0])
    AtAx = parallel_model.sparse_back_project(Ax, indices[0]).reshape(x.shape)
    finite_diff_hessian = AtAx[i, j, k] / eps
    print('Hessian matches finite difference: {}'.format(jnp.allclose(hessian.reshape(x.shape)[i, j, k], finite_diff_hessian)))

    # ##########################
    # Check the time taken per forward projection
    #  NOTE: recompiling happens whenever sparse_forward_project is called with a new *length* of input indices
    time_taken = 0

    print('\nStarting multiple forward projections...')
    for j in range(num_trials):
        voxel_values = x.reshape((-1, num_recon_slices))[indices[j]]
        t0 = time.time()
        fp = parallel_model.sparse_forward_project(voxel_values, indices[j])
        time_taken += time.time() - t0
        del fp
        gc.collect()

    print('Mean time per call = {}'.format(time_taken / num_trials))
    print('Done')

    # ##########################
    # Check the time taken per backprojection
    #  NOTE: recompiling happens whenever sparse_back_project is called with a new *length* of input indices
    time_taken = 0

    print('\nStarting multiple backprojections...')
    for j in range(num_trials):
        t0 = time.time()
        bp = parallel_model.sparse_back_project(sinogram, indices[j])
        time_taken += time.time() - t0
        del bp
        gc.collect()

    print('Mean time per call = {}'.format(time_taken / num_trials))
    print('Done')

    # ##########################
    # Show the forward and back projection from a single pixel
    i, j = num_recon_rows // 4, num_recon_cols // 3
    x = jnp.zeros(recon_shape)
    x = x.at[i, j, :].set(1)
    voxel_values = x.reshape((-1, num_recon_slices))[indices[0]]

    Ax = parallel_model.sparse_forward_project(voxel_values, indices[0])
    Aty = parallel_model.sparse_back_project(Ax, indices[0])
    Aty = parallel_model.reshape_recon(Aty)

    y = jnp.zeros_like(sinogram)
    view_index = 30
    y = y.at[view_index].set(sinogram[view_index])
    index = jnp.ravel_multi_index((60, 60), (num_recon_rows, num_recon_cols))
    a1 = parallel_model.sparse_back_project(y, indices[0])

    slice_index = 0
    fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
    ax[0].imshow(x[:, :, slice_index])
    ax[0].set_title('x = phantom')
    ax[1].imshow(Ax[:, slice_index, :])
    ax[1].set_title('y = Ax')
    ax[2].imshow(Aty[:, :, slice_index])
    ax[2].set_title('Aty = AtAx')
    plt.pause(2)

    print('Final memory stats:')
    mj.get_memory_stats(print_results=True)
    input('Press return to exit')
    a = 0



import warnings
import jax
import jax.numpy as jnp
from functools import partial
from collections import namedtuple

import numpy as np

import mbirjax


class TranslationModel(mbirjax.TomographyModel):
    """
    Implements a tomography model using translation-based cone-beam projections.

    This class inherits from `mbirjax.TomographyModel` and specializes it for scenarios where the object undergoes
    translations instead of rotations during image acquisition. The forward and backward projections are computed using
    per-view translation vectors.

    Note: This geometry reduces regularization along the row (y) direction by setting `qggmrf_nbr_wts=(0.5, 1, 1)`,
    which helps compensates for anisotropy intrinsic to the translation geometry.

    Args:
        sinogram_shape (tuple): Shape of the sinogram as (num_views, num_rows, num_channels),
            where 'num_views' is the number of translation steps, 'num_rows' is the number of detector rows,
            and 'num_channels' is the number of detector columns.
        translation_vectors (jax array or numpy array): A (num_views, 3) array of translations (x, y, z) in ALUs.
            Each vector specifies how the object is translated for each view.
            Positive x shifts the object left, z shifts up, and y shifts away from the source.
        source_detector_dist (float): Distance from the X-ray source to the detector.
        source_iso_dist (float): Distance from the X-ray source to the isocenter.


    Usage Notes:
        - Reconstruction parameters should typically be set using `set_params`.
        - The row voxel thickness can be adjusted with `delta_recon_row`, while the voxel column and slice spacing
          can be controlled via the base parameter `delta_voxel`.

    See Also:
        mbirjax.TomographyModel: Base class with standard methods like `set_params` and `reconstruct`.

    Example:
        .. code-block:: python

            import jax.numpy as jnp
            from mbirjax.translation_model import TranslationModel

            sinogram_shape = (180, 256, 256)
            translation_vectors = jnp.zeros((180, 3))
            model = TranslationModel(
                sinogram_shape=sinogram_shape,
                translation_vectors=translation_vectors,
                source_detector_dist=500.0,
                source_iso_dist=500.0
            )
            model.set_params(delta_recon_row=2.0)
            model.auto_set_recon_shape(sinogram_shape)
    """
    def __init__(self, sinogram_shape, translation_vectors, source_detector_dist, source_iso_dist):

        self.use_ror_mask = False
        self.entries_per_cylinder_batch = 128
        translation_vectors = jnp.asarray(translation_vectors)
        view_params_name = 'translation_vectors'

        super().__init__(sinogram_shape, translation_vectors=translation_vectors, source_detector_dist=source_detector_dist,
                         source_iso_dist=source_iso_dist, view_params_name=view_params_name, qggmrf_nbr_wts=(0.1, 1, 1,))

    def get_magnification(self):
        """
        Returns the magnification for the cone beam geometry.

        Returns:
            magnification = source_detector_dist / source_iso_dist
        """
        source_detector_dist, source_iso_dist = self.get_params(['source_detector_dist', 'source_iso_dist'])
        if jnp.isinf(source_detector_dist):
            raise ValueError('Distance from source to detector is infinite, which means all translated projections have the same information.')
        else:
            magnification = source_detector_dist / source_iso_dist
        return magnification

    def get_geometry_parameters(self):
        """
        Function to get a list of the primary geometry parameters for translation model projection.

        Returns:
            namedtuple of required geometry parameters.
        """
        # First get the parameters managed by ParameterHandler
        geometry_param_names = \
            ['delta_det_row', 'delta_det_channel', 'det_row_offset', 'det_channel_offset',
             'source_detector_dist', 'delta_voxel', 'delta_recon_row']
        geometry_param_values = self.get_params(geometry_param_names)

        # Then get additional parameters:
        geometry_param_names += ['magnification', 'psf_radius',
                                 'entries_per_cylinder_batch']
        geometry_param_values.append(self.get_magnification())
        geometry_param_values.append(self.get_psf_radius())
        geometry_param_values.append(self.entries_per_cylinder_batch)

        # Then create a namedtuple to access parameters by name in a way that can be jit-compiled.
        GeometryParams = namedtuple('GeometryParams', geometry_param_names)
        geometry_params = GeometryParams(*tuple(geometry_param_values))

        return geometry_params

    def get_psf_radius(self):
        """
        Compute the integer radius of the PSF kernel for cone beam projection.
        """
        delta_det_row, delta_det_channel, source_iso_dist, source_detector_dist, recon_shape, delta_voxel, delta_recon_row, translation_vectors = self.get_params(
            ['delta_det_row', 'delta_det_channel', 'source_iso_dist', 'source_detector_dist', 'recon_shape', 'delta_voxel', 'delta_recon_row', 'translation_vectors'])
        magnification = self.get_magnification()

        # Compute minimum detector pitch
        delta_det = jnp.minimum(delta_det_row, delta_det_channel)

        # Find the maximum and minimum translation vectors
        max_translation = jnp.amax(translation_vectors, axis=0)
        min_translation = jnp.amin(translation_vectors, axis=0)

        # Compute maximum magnification
        if jnp.isinf(source_detector_dist):
            raise ValueError('Distance from source to detector is infinite, which means all translated projections have the same information.')
        else:
            # Determine the distance from the source to the closest voxel.
            source_to_closest_pixel = source_iso_dist - (0.5 * recon_shape[0] * delta_recon_row) - max_translation[1]
            # Determine the maximum magnification.
            max_magnification = source_detector_dist / source_to_closest_pixel

            if source_to_closest_pixel < 0:
                raise ValueError('Reconstruction volume extends into source - no valid projection in this case.')

        # Compute the maximum number of detector rows/channels on either side of the center detector hit by a voxel
        psf_radius = int(jnp.ceil(jnp.ceil((delta_voxel * max_magnification / delta_det)) / 2))
        if psf_radius > 4:
            warnings.warn('A single voxel may project onto 100 or more detector elements, which may lead to artifacts. Consider using smaller voxels.')
        return psf_radius

    def auto_set_recon_shape(self, sinogram_shape, no_compile=True, no_warning=False):
        """ Compute the automatic recon shape translation reconstruction.
        """
        # Get model parameters
        source_detector_dist, source_iso_dist = self.get_params(['source_detector_dist', 'source_iso_dist'])
        delta_det_row, delta_det_channel = self.get_params(['delta_det_row', 'delta_det_channel'])
        magnification = self.get_magnification()
        delta_voxel = self.get_params('delta_voxel')
        num_views, num_det_rows, num_det_channels = sinogram_shape
        translation_vectors = self.get_params('translation_vectors')

        # Calculate the width and height of the detector in ALU
        detect_box = jnp.array([delta_det_channel*num_det_channels, delta_det_row*num_det_rows])

        # Compute cone_slope = tan(cone_angle/2) along the x and z directions
        cone_slope = (detect_box/2) / source_detector_dist

        # Compute detector pixel pitch at iso
        # Note that this may differ from delta_voxel
        # However, we will use det_pixel_pitch_iso to calculate both the number rows and their pitch
        det_pixel_pitch_iso_vec = jnp.array([delta_det_row, delta_det_channel])/magnification
        det_pixel_pitch_iso = jnp.max(det_pixel_pitch_iso_vec)

        # Compute the row pitch based on a heuristic
        # This results in isotrqopic voxels when cone_angle/2 > 63 deg
        #ToDo: There will be problems if cone_slope is small or zero. Discuss with Greg.
        nominal_row_pitch = 2.0*det_pixel_pitch_iso_vec/cone_slope
        nominal_row_pitch = jnp.max(nominal_row_pitch)  # Take the maximum of the nominal pitches along x and z
        delta_recon_row = jnp.maximum(nominal_row_pitch, det_pixel_pitch_iso) # Ensure that the row resolution is not higher than the (x,z) detector resolution
        delta_recon_row = float(delta_recon_row)

        # Compute cube = (width, depth, height) of the scanned region in ALU
        max_translation = jnp.amax(translation_vectors, axis=0)  # Translate object right/up when positive
        min_translation = jnp.amin(translation_vectors, axis=0)  # Translate object left/down when negative
        cube = max_translation - min_translation

        # Compute recon_box = (width, height) of the reconstruction box in "nominal voxels"
        # Nominal voxels are the voxels that would result using the detector pitch at iso
        recon_box = jnp.ceil(jnp.array([cube[0], cube[2]]) / det_pixel_pitch_iso_vec)

        # ************ Use a heuristic to determine a reasonable number of rows *************
        # Compute the number of unknown pixels per view
        num_pixels_per_view = ((recon_box[0] + num_det_rows)*(recon_box[1] + num_det_channels)) / num_views
        num_measurements_per_view = num_det_channels*num_det_rows
        # Select the number of rows to match the number of unknowns to the number of measurements
        num_recon_rows = jnp.ceil( num_measurements_per_view / num_pixels_per_view )

        # Make sure the object extends no further than halfway to the source
        max_recon_rows = jnp.floor((source_iso_dist - cube[1]) / delta_recon_row)
        if max_recon_rows < 1:
            print(f"[Error] Computed max_recon_rows = {max_recon_rows} < 1. This suggests the object extends beyond the source.")
        num_recon_rows = jnp.minimum(num_recon_rows, max_recon_rows)

        # Set the parameters to their computed values
        num_recon_cols, num_recon_slices = recon_box
        num_recon_cols = int(num_recon_cols)
        num_recon_rows = int(num_recon_rows)
        num_recon_slices = int(num_recon_slices)
        recon_shape = (num_recon_rows, num_recon_cols, num_recon_slices)
        self.set_params(no_compile=no_compile, no_warning=no_warning, recon_shape=recon_shape, delta_recon_row=delta_recon_row)


    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def forward_project_pixel_batch_to_one_view(voxel_values, pixel_indices, translation_vector, projector_params):
        """
        Forward project a set of voxels determined by indices into the flattened array of size num_rows x num_cols.

        Args:
            voxel_values (jax array):  2D array of shape (num_indices, num_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of indices into flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows, num_det_channels)
        """
        recon_shape = projector_params.recon_shape
        num_recon_slices = recon_shape[2]
        if voxel_values.shape[0] != pixel_indices.shape[0] or len(voxel_values.shape) < 2 or \
                voxel_values.shape[1] != num_recon_slices:
            raise ValueError('voxel_values must have shape[0:2] = (num_indices, num_slices)')

        vertical_fan_projector = TranslationModel.forward_vertical_fan_pixel_batch_to_one_view
        horizontal_fan_projector = TranslationModel.forward_horizontal_fan_pixel_batch_to_one_view

        new_voxel_values = vertical_fan_projector(voxel_values, pixel_indices, translation_vector, projector_params)
        sinogram_view = horizontal_fan_projector(new_voxel_values, pixel_indices, translation_vector, projector_params)

        return sinogram_view

    @staticmethod
    def forward_vertical_fan_pixel_batch_to_one_view(voxel_values, pixel_indices, translation_vector, projector_params):
        """
        Apply a fan beam forward projection in the vertical direction separately to each voxel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns an array corresponding to the same pixel
        locations, but using values obtained from the projection of the original voxel cylinders onto a detector column,
        so the output array has size (len(pixel_indices), num_det_rows).

        Args:
            voxel_values (jax array):  2D array of shape (num_pixels, num_recon_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of shape (len(pixel_indices), ) holding the indices into
                the flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_pixels, num_det_rows)
        """
        pixel_map = jax.vmap(TranslationModel.forward_vertical_fan_one_pixel_to_one_view,
                             in_axes=(0, 0, None, None))
        new_pixels = pixel_map(voxel_values, pixel_indices, translation_vector, projector_params)

        return new_pixels

    @staticmethod
    def forward_horizontal_fan_pixel_batch_to_one_view(voxel_values, pixel_indices, translation_vector, projector_params):
        """
        Apply a horizontal fan beam transformation to a set of voxel cylinders. These cylinders are assumed to have
        slices aligned with detector rows, so that a horizontal fan beam maps a cylinder slice to a detector row.
        This function returns the resulting sinogram view.

        Args:
            voxel_values (jax array):  2D array of shape (num_pixels, num_recon_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of shape (len(pixel_indices), ) holding the indices into
                the flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows, num_det_channels)
        """

        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape

        # Get the data needed for horizontal projection
        n_p, n_p_center, W_p_c, cos_theta_p = TranslationModel.compute_horizontal_data(pixel_indices, translation_vector, projector_params)
        L_max = jnp.minimum(1, W_p_c)

        # Allocate the sinogram array
        sinogram_view = jnp.zeros((num_det_rows, num_det_channels))

        # Do the horizontal projection
        for n_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
            n = n_p_center + n_offset
            abs_delta_p_c_n = jnp.abs(n_p - n)
            L_p_c_n = jnp.clip((W_p_c + 1) / 2 - abs_delta_p_c_n, 0, L_max)
            A_chan_n = L_p_c_n / cos_theta_p
            A_chan_n *= (n >= 0) * (n < num_det_channels)
            sinogram_view = sinogram_view.at[:, n].add(A_chan_n.reshape((1, -1)) * voxel_values.T)

        return sinogram_view

    @staticmethod
    def forward_vertical_fan_one_pixel_to_one_view(voxel_cylinder, pixel_index, translation_vector, projector_params):
        """
        Apply a fan beam forward projection in the vertical direction to the pixel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns a vector obtained from the projection of
        the original voxel cylinder onto a detector column, so the output vector has length num_det_rows.

        Args:
            voxel_cylinder (jax array):  1D array of shape (num_recon_slices, ) of voxel values, where
                voxel_cylinder[j] is the value of the voxel in slice j at the location determined by pixel_index.
            pixel_index (int):  Index into the flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows,)

        Note:
            This is a helper function used in vmap in :meth:`TranslationModel.forward_vertical_fan_pixel_batch_to_one_view`
        This method has the same signature and output as that method, except single int pixel_index is used
        in place of the 1D pixel_indices, and likewise only a single voxel cylinder is returned.
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape
        num_slices = voxel_cylinder.shape[0]

        # From pixel index, compute y and pixel_mag
        y, pixel_mag = TranslationModel.compute_y_mag_for_pixel(pixel_index, translation_vector, recon_shape,
                                                                projector_params)

        # The code above depends only on the pixel - a single point.  z is a potentially large vector
        # Here we compute cos_phi_p:  1 / cos_phi_p determines the projection length through a voxel
        # For computational efficiency, we use that to scale the voxel_cylinder values.
        ### possibly convert to a jitted function with donate_argnames to avoid copies for z, v, phi_p, cos_phi_p
        k = jnp.arange(len(voxel_cylinder))
        z = gp.delta_voxel * (k - (num_recon_slices - 1) / 2.0) - translation_vector[2] # recon_ijk_to_xyz
        v = pixel_mag * z  # geometry_xyz_to_uv_mag
        # Compute vertical cone angle of voxels
        phi_p = jnp.arctan2(v, gp.source_detector_dist)  # compute_vertical_data_single_pixel
        cos_phi_p = jnp.cos(phi_p)  # We assume the vertical angle |phi_p| < 45 degrees so cos_alpha_p_z = cos_phi_p
        scaled_voxel_values = voxel_cylinder / cos_phi_p
        # End

        # Get the length of projection of detector on vertical voxel profile (in fraction of voxel size)
        # This is also the slope of the map from voxel index to detector index
        W_p_r = (pixel_mag * gp.delta_voxel) / gp.delta_det_row
        slope_k_to_m = W_p_r
        L_max = jnp.minimum(1, W_p_r)  # Maximum fraction of a detector that can be covered by one voxel.

        # Set up detector row indices array (0, 10, 20, ..., 10*num_slice_batches)
        det_rows_per_batch = gp.entries_per_cylinder_batch
        det_rows_per_batch = min(det_rows_per_batch, num_det_rows)
        num_det_row_batches = (num_det_rows + det_rows_per_batch - 1) // det_rows_per_batch
        det_row_indices = det_rows_per_batch * jnp.arange(num_det_row_batches)

        det_center_row = (num_det_rows - 1) / 2.0
        row_batch = jnp.arange(det_rows_per_batch)

        # Set up a function to map over subsets of detector rows
        def create_det_column_rows(start_index):
            # We need to match the back projector, so we have to determine the fraction of each voxel that projects
            # to each detector.
            # First project the detector centers to the voxel cylinder
            m_center = start_index + row_batch  # Center of detector elements
            v_m = (m_center - det_center_row) * gp.delta_det_row - gp.det_row_offset  # Detector center in ALUs
            z_m = v_m / pixel_mag  # z coordinate of the projection of the center of the first detector element in this batch
            # Convert to voxel fractional index and find the center of each voxel
            k_m = (z_m + translation_vector[2]) / gp.delta_voxel + (num_slices - 1) / 2.0
            k_m_center = jnp.round(k_m).astype(int)  # Center of the voxel hit by the center of the detector
            # Then map the center of the voxels back to the detector.
            m_p = slope_k_to_m * (k_m_center - k_m[0]) + m_center[0]  # Projection to detector of voxel centers

            # Allocate space
            new_column_batch = jnp.zeros(det_rows_per_batch)
            # Do the vertical projection
            for k_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
                k_ind = k_m_center + k_offset  # Indices of the current set of voxels touched by the detector elements
                # The projection of these centers is the projection of k_m_center (which is m_p) plus
                # the offset times the slope of the map from voxel index to detector index
                abs_delta_p_r_m = jnp.abs(m_p + slope_k_to_m * k_offset - m_center)  # Distance from projection of center of voxel to center of detector
                A_row_k = jnp.clip((W_p_r + 1) / 2 - abs_delta_p_r_m, 0, L_max)  # Fraction of the detector hit by this voxel
                A_row_k *= (k_ind >= 0) * (k_ind < num_slices)
                new_column_batch = jnp.add(new_column_batch, A_row_k * scaled_voxel_values[k_ind])

            return new_column_batch, None

        det_column, _ = jax.lax.map(create_det_column_rows, det_row_indices)
        det_column = det_column.flatten()
        det_column = jax.lax.slice_in_dim(det_column, 0, num_det_rows)
        return det_column

    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def back_project_one_view_to_pixel_batch(sinogram_view, pixel_indices, single_view_params, projector_params,
                                             coeff_power=1):
        """
        Use vmap to do a backprojection from one view to multiple pixels (voxel cylinders).

        Args:
            sinogram_view (2D jax array): one view of the sinogram to be back projected.
                2D jax array of shape (num_det_rows)x(num_det_channels)
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for the view being back projected.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            The voxel values for all slices at the input index (i.e., a voxel cylinder) obtained by backprojecting
            the input sinogram view.
        """

        vertical_fan_projector = TranslationModel.back_vertical_fan_one_view_to_pixel_batch
        horizontal_fan_projector = TranslationModel.back_horizontal_fan_one_view_to_pixel_batch

        det_voxel_cylinder = horizontal_fan_projector(sinogram_view, pixel_indices, single_view_params,
                                                      projector_params, coeff_power=coeff_power)
        back_projection = vertical_fan_projector(det_voxel_cylinder, pixel_indices, single_view_params,
                                                 projector_params, coeff_power=coeff_power)

        return back_projection

    @staticmethod
    def back_horizontal_fan_one_view_to_pixel_batch(sinogram_view, pixel_indices, translation_vector,
                                                    projector_params, coeff_power=1):
        """
        Apply the back projection of a horizontal fan beam transformation to a single sinogram view
        and return the resulting voxel cylinders.

        Args:
            sinogram_view (2D jax array): one view of the sinogram to be back projected.
                2D jax array of shape (num_det_rows)x(num_det_channels)
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.
        Returns:
            jax array of shape (len(pixel_indices), num_det_rows)
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        num_pixels = pixel_indices.shape[0]

        # Get the data needed for horizontal projection
        n_p, n_p_center, W_p_c, cos_theta_p = TranslationModel.compute_horizontal_data(pixel_indices, translation_vector, projector_params)
        L_max = jnp.minimum(1, W_p_c)

        # Allocate the voxel cylinder array
        det_voxel_cylinder = jnp.zeros((num_pixels, num_det_rows))

        # Do the horizontal projection
        for n_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
            n = n_p_center + n_offset
            abs_delta_p_c_n = jnp.abs(n_p - n)
            L_p_c_n = jnp.clip((W_p_c + 1) / 2 - abs_delta_p_c_n, 0, L_max)
            A_chan_n = L_p_c_n / cos_theta_p
            A_chan_n *= (n >= 0) * (n < num_det_channels)
            A_chan_n = A_chan_n ** coeff_power
            det_voxel_cylinder = jnp.add(det_voxel_cylinder, A_chan_n.reshape((-1, 1)) * sinogram_view[:, n].T)

        return det_voxel_cylinder

    @staticmethod
    def back_vertical_fan_one_view_to_pixel_batch(det_voxel_cylinder, pixel_indices, single_view_params,
                                                  projector_params, coeff_power=1):
        """
        Apply a fan beam backward projection in the vertical direction to the pixel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns a vector obtained from the projection of
        the detector-based voxel cylinders onto voxel cylinders in recon space, so the output vector has length
        num_recon_slices.

        Args:
            det_voxel_cylinder (2D jax array): 2D array of shape (num_pixels, num_det_rows) of voxel values, where
                det_voxel_cylinder[i, j] is the value of the voxel in row j at the location determined by indices[i].
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for the view being back projected.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            2D jax array of shape (num_pixels, num_recon_slices) of voxel values.
        """
        pixel_map = jax.vmap(TranslationModel.back_vertical_fan_one_view_to_one_pixel,
                             in_axes=(0, 0, None, None, None))
        new_pixels = pixel_map(det_voxel_cylinder, pixel_indices, single_view_params, projector_params, coeff_power)

        return new_pixels

    @staticmethod
    def back_vertical_fan_one_view_to_one_pixel(detector_column_values, pixel_index, translation_vector, projector_params,
                                                coeff_power=1):
        """
        Apply the back projection of a vertical fan beam transformation to a single voxel cylinder and return the column
        vector of the resulting values.

        Args:
            detector_column_values (1D jax array): 1D array of shape (num_det_rows,) of voxel values, where
                detector_column_values[i, j] is the value of the voxel in row j at the location determined by indices[i].
            pixel_index (int):  Index into flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            1D jax array of shape (num_recon_slices,) of voxel values.
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Set up slice indices array (0, slices_per_batch, 2*slices_per_batch, ..., num_slice_batches*slices_per_batch)
        num_slices = num_recon_slices
        slices_per_batch = gp.entries_per_cylinder_batch
        slices_per_batch = min(slices_per_batch, num_slices)
        num_slice_batches = (num_slices + slices_per_batch - 1) // slices_per_batch
        slice_indices = slices_per_batch * jnp.arange(num_slice_batches)

        # Set up a function to map over the slices of the cylinder
        # Here we can use a map over subsections of the voxel cylinder because we are indexing by slice,
        # so there is no overlap from one section to the next.
        def create_voxel_cylinder_slices(start_index):
            # Allocate space
            new_cylinder = jnp.zeros(slices_per_batch)
            # Get the data needed for vertical projection
            cur_slice_indices = start_index + jnp.arange(slices_per_batch)
            m_p, m_p_center, W_p_r, cos_alpha_p_z = TranslationModel.compute_vertical_data_single_pixel(pixel_index, cur_slice_indices, translation_vector, projector_params)
            L_max = jnp.minimum(1, W_p_r)  # Maximum fraction of a detector that can be covered by one voxel.

            # Do the vertical projection
            for m_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
                m = m_p_center + m_offset
                abs_delta_p_r_m = jnp.abs(m_p - m)  # Distance from projection of center of voxel to center of detector
                L_p_r_m = jnp.clip((W_p_r + 1) / 2 - abs_delta_p_r_m, 0, L_max)
                A_row_m = L_p_r_m / cos_alpha_p_z
                A_row_m *= (m >= 0) * (m < num_det_rows)
                A_row_m = A_row_m ** coeff_power
                new_cylinder = jnp.add(new_cylinder, A_row_m * detector_column_values[m])

            return new_cylinder, None

        recon_voxel_cylinder, _ = jax.lax.map(create_voxel_cylinder_slices, slice_indices)
        recon_voxel_cylinder = recon_voxel_cylinder.flatten()
        recon_voxel_cylinder = jax.lax.slice_in_dim(recon_voxel_cylinder, 0, num_recon_slices)
        return recon_voxel_cylinder

    @staticmethod
    def compute_vertical_data_single_pixel(pixel_index, slice_indices, translation_vector, projector_params):
        """
        Compute the quantities m_p, m_p_center, W_p_r, cos_alpha_p_z needed for vertical projection.

        Args:
            pixel_index (int):  Index into flattened array of size num_rows x num_cols.
            slice_indices (array of int): Indices into the recon slices.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).

        Returns:
            m_p, m_p_center, W_p_r, cos_alpha_p_z
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Convert the index into (i,j,k) coordinates corresponding to the indices into the 3D voxel array
        row_index, col_index = jnp.unravel_index(pixel_index, recon_shape[:2])
        # slice_indices = jnp.arange(num_recon_slices)

        x_p, y_p, z_p = TranslationModel.recon_ijk_to_xyz(row_index, col_index, slice_indices, recon_shape, gp.delta_voxel, gp.delta_recon_row, translation_vector)

        # Convert from xyz to coordinates on detector
        u_p, v_p, pixel_mag = TranslationModel.geometry_xyz_to_uv_mag(x_p, y_p, z_p, gp.source_detector_dist, gp.magnification)

        # Convert from uv to index coordinates in detector and get the vector of center detector rows for this cylinder
        m_p, _ = TranslationModel.detector_uv_to_mn(u_p, v_p, gp.delta_det_channel, gp.delta_det_row, gp.det_channel_offset,
                                                    gp.det_row_offset, num_det_rows, num_det_channels)

        m_p_center = jnp.round(m_p).astype(int)

        # Compute vertical cone angle of pixel
        phi_p = jnp.arctan2(v_p, gp.source_detector_dist)

        # Compute cos alpha for row and columns
        cos_phi_p = jnp.cos(phi_p)  # We assume the vertical angle |phi_p| < 45 degrees so cos_alpha_p_z = cos_phi_p
        # cos_alpha_p_z = jnp.maximum(jnp.abs(cos_phi_p), jnp.abs(jnp.sin(phi_p)))

        # Get the length of projection of flattened voxel on detector (in fraction of detector size)
        W_p_r = pixel_mag * (gp.delta_voxel / gp.delta_det_row)  # * cos_alpha_p_z / cos_phi_p

        vertical_data = (m_p, m_p_center, W_p_r, cos_phi_p)  # cos_alpha_p_z)

        return vertical_data

    @staticmethod
    def compute_horizontal_data(pixel_indices, translation_vector, projector_params):
        """
        Compute the quantities n_p, n_p_center, W_p_c, cos_alpha_p_xy needed for vertical projection.

        Args:
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            translation_vector (jax array of floats):  1D translation vector in ALU units for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).

        Returns:
            n_p, n_p_center, W_p_c, cos_theta_p
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Convert the index into (i,j,k) coordinates corresponding to the indices into the 3D voxel array
        row_index, col_index = jnp.unravel_index(pixel_indices, recon_shape[:2])
        slice_index = jnp.arange(1)

        x_p, y_p, _ = TranslationModel.recon_ijk_to_xyz(row_index, col_index, slice_index, recon_shape, gp.delta_voxel, gp.delta_recon_row, translation_vector)

        # Convert from xyz to coordinates on detector
        # pixel_mag should be kept in terms of magnification to allow for source_detector_dist = jnp.Inf
        pixel_mag = 1 / (1 / gp.magnification - y_p / gp.source_detector_dist)
        # Compute the physical position that this voxel projects onto the detector
        u_p = pixel_mag * x_p
        det_center_channel = (num_det_channels - 1) / 2.0  # num_of_cols

        # Calculate indices on the detector grid
        n_p = (u_p + gp.det_channel_offset) / gp.delta_det_channel + det_center_channel  # Sync with detector_uv_to_mn
        n_p_center = jnp.round(n_p).astype(int)

        # Compute horizontal and vertical cone angle of pixel
        theta_p = jnp.arctan2(u_p, gp.source_detector_dist)

        # Compute cosine alpha
        cos_theta_p = jnp.cos(theta_p)

        # Compute projected voxel width along columns and rows (in fraction of detector size)
        W_p_c = pixel_mag * (gp.delta_voxel / gp.delta_det_channel)

        horizontal_data = (n_p, n_p_center, W_p_c, cos_theta_p)

        return horizontal_data

    @staticmethod
    def recon_ijk_to_xyz(i, j, k, recon_shape, delta_voxel, delta_recon_row, translation_vector):
        """
        Convert (i, j, k) indices into the recon volume to corresponding (x, y, z) coordinates.
        """
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        y = delta_recon_row * (i - (num_recon_rows - 1) / 2.0) - translation_vector[1]
        x = delta_voxel * (j - (num_recon_cols - 1) / 2.0) - translation_vector[0]
        z = delta_voxel * (k - (num_recon_slices - 1) / 2.0) - translation_vector[2]
        return x, y, z

    @staticmethod
    def geometry_xyz_to_uv_mag(x, y, z, source_detector_dist, magnification):
        """
        Convert (x, y, z) coordinates to to (u, v) detector coordinates plus the pixel-dependent magnification.
        """
        # Compute the magnification at this specific voxel
        # The following expression is valid even when source_detector_dist = jnp.Inf
        pixel_mag = 1 / (1 / magnification - y / source_detector_dist)

        # Compute the physical position that this voxel projects onto the detector
        u = pixel_mag * x
        v = pixel_mag * z

        return u, v, pixel_mag

    @staticmethod
    @jax.jit
    def detector_uv_to_mn(u, v, delta_det_channel, delta_det_row, det_channel_offset, det_row_offset, num_det_rows,
                          num_det_channels):
        """
        Convert (u, v) detector coordinates to fractional indices (m, n) into the detector.

        Note:
            This version does not account for nonzero detector rotation.
        """
        # Get the center of the detector grid for columns and rows
        det_center_row = (num_det_rows - 1) / 2.0  # num_of_rows
        det_center_channel = (num_det_channels - 1) / 2.0  # num_of_cols

        # Calculate indices on the detector grid
        m = (v + det_row_offset) / delta_det_row + det_center_row
        n = (u + det_channel_offset) / delta_det_channel + det_center_channel  # Sync with compute_horizontal_data

        return m, n

    @staticmethod
    @jax.jit
    def compute_y_mag_for_pixel(pixel_index, translation_vector, recon_shape, projector_params):

        gp = projector_params.geometry_params
        row_index, col_index = jnp.unravel_index(pixel_index, recon_shape[:2])
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        y = gp.delta_recon_row * (row_index - (num_recon_rows - 1) / 2.0) - translation_vector[1]

        # Convert from xyz to coordinates on detector
        pixel_mag = 1 / (1 / gp.magnification - y / gp.source_detector_dist)
        return y, pixel_mag


